"""
Module defining objects for TC-related conventions
"""

import numpy as np
from metpy.units import units

# %% Categories
_thresholds = {
    "Klotzbach": dict(
        bins=np.asarray([-np.inf, 925, 945, 960, 975, 990, 1005, np.inf])
        * units("hPa"),
        labels=[5, 4, 3, 2, 1, 0, -1],
    ),
    "Simpson": dict(
        bins=np.asarray([-np.inf, 920, 945, 965, 970, 980, 990, np.inf]) * units("hPa"),
        labels=[5, 4, 3, 2, 1, 0, -1],
    ),
    "Saffir-Simpson": dict(
        bins=np.asarray([-np.inf, 16, 29, 38, 44, 52, 63, np.inf]) * units("m s-1"),
        labels=[-1, 0, 1, 2, 3, 4, 5],
    ),
}



import gzip

import datetime
import warnings

import cftime
import numpy as np
import xarray as xr

from parse import parse

# Use align specifications (^, <, >) to allow variable whitespace in headers
# Left aligned (<) for "nvars" so nfields takes all whitespace between in case there is
# only one space
# Right aligned (>) for variables at the end since lines are stripped of whitespace
# prior to parsing
header_fmt = "TRACK_NUM{ntracks:^d}ADD_FLD{nfields:^d}{nvars:<d}&{var_has_coords}"
track_header_fmt = "TRACK_ID{track_id:>d}"
track_header_fmt_new = "TRACK_ID{track_id:^d}START_TIME{start_time:>}"
track_info_fmt = "POINT_NUM{npoints:>d}"

tilt_header_fmt = "NTRACK {ntracks:d} NFIELD {nfields:d}"
tilt_track_header_fmt = "TRACK_NO {track_id:d} NUMPT {npoints:d}"


def _parse(fmt, string, **kwargs):
    # Call parse but raise an error if None is returned
    result = parse(fmt, string, **kwargs)

    if result is None:
        raise ValueError(f"Format {fmt} does not match string {string}")

    return result


def parse_date(date, calendar=None):
    if isinstance(calendar, (tuple, list)):
        initial_date = np.datetime64(calendar[0])
        timestep = calendar[1]
        if not isinstance(timestep, np.timedelta64):
            timestep = np.timedelta64(timestep, "h")
        return initial_date + (int(date) - 1) * timestep

    elif len(date) == 10:  # i.e., YYYYMMDDHH
        if calendar is not None:
            yr = int(date[0:4])
            mn = int(date[4:6])
            dy = int(date[6:8])
            hr = int(date[8:10])
            return cftime.datetime(yr, mn, dy, hr, calendar=calendar)
        else:
            return datetime.datetime.strptime(date.strip(), "%Y%m%d%H")

    else:
        return int(date)


def load(filename, calendar=None, variable_names=None):
    """Load ASCII TRACK data as an xarray.Dataset

    Parameters
    ----------
    filename: str
    calendar : optional
    variable_names : list of str, optional
        TRACK

    Returns
    -------
    xarray.Dataset
    """
    if filename.split(".")[-1] == "gz":
        open_func = gzip.open
    else:
        open_func = open

    output = list()
    track_id = list()
    times = list()
    with open_func(filename, "rt") as f:
        # The first lines can contain extra information bounded by two extra lines
        # Just skip to the main header line for now
        line = ""
        while not line.startswith("TRACK_NUM"):
            line = f.readline().strip()

        # Load information about tracks from header line
        # If there are no added variables the line ends at the "&"
        try:
            header = _parse(header_fmt, line).named
        except ValueError:
            header = _parse(header_fmt.split("&")[0] + "&", line).named
            header["var_has_coords"] = ""

        ntracks = header["ntracks"]
        nfields = header["nfields"]
        nvars = header["nvars"]
        has_coords = [int(x) == 1 for x in header["var_has_coords"]]

        # Check header data is consistent
        if len(has_coords) != nfields:
            raise ValueError(
                f"TRACK file header is inconsistent. Number of fields "
                f"({len(has_coords)}) does not match nfields from header ({nfields})."
            )

        nvars_expected = sum([3 if x == 1 else 1 for x in has_coords])
        if nvars_expected != nvars:
            raise ValueError(
                f"TRACK file header is inconsistent. Number of variables including"
                f"lat/lon for variables ({nvars_expected}) does not match nvars from"
                f"header ({nvars})."
            )

        # Create a list of variables stored in each track
        # Generic names for variables as there is currently no information otherwise
        var_labels = ["lon", "lat", "vorticity"]
        if variable_names is None:
            variable_names = [f"feature_{n}" for n in range(nfields)]
        else:
            if len(variable_names) != nfields:
                raise ValueError(
                    f"Number of variable names given ({len(variable_names)}) does not"
                    f"match number of fields in file ({len(nfields)})"
                )
        for n, variable_name in enumerate(variable_names):
            if has_coords[n]:
                var_labels.append(f"{variable_name}_lon")
                var_labels.append(f"{variable_name}_lat")
            var_labels.append(variable_name)

        # Read in each track as an xarray dataset with time as the coordinate
        for n in range(ntracks):
            # Read individual track header (two lines)
            line = f.readline().strip()
            if not line.replace(" ", "") == "":  # If line is empty
                try:
                    track_info = _parse(track_header_fmt, line).named
                except ValueError:
                    track_info = _parse(track_header_fmt_new, line).named

                line = f.readline().strip()
                npoints = _parse(track_info_fmt, line)["npoints"]

                # Generate arrays for time coordinate and variables
                # Time is a list because it will hold datetime or cftime objects
                # Other variables are a dictionary mapping variable name to a tuple of
                # (time, data_array) as this is what is passed to xarray.Dataset

                # Add track ID as a variable along the record dimension so that it can
                # be used for groupby
                track_id.extend([track_info["track_id"]] * npoints)

                # Populate time and data line by line
                for m in range(npoints):
                    line = f.readline().strip()
                    times.append(
                        parse_date(
                            line.split(" ")[0],
                            calendar=calendar,
                        )
                    )
                    output.append(line.replace("&", " "))
            else:
                warnings.warn(
                    "Parsed line is empty. It is possible this problem arrises because"
                    "the number of tracks expected from the header was not found in the"
                    "file."
                )

    output = np.genfromtxt(output)
    data_vars = {
        varname: ("record", output[:, i])
        for i, varname in enumerate(var_labels, start=1)
    }
    data_vars["track_id"] = ("record", track_id)
    data_vars["time"] = ("record", times)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message="Converting non-nanosecond precision datetime values to nanosecond",
        )
        ds = xr.Dataset(data_vars=data_vars)
    ds.track_id.attrs["cf_role"] = "trajectory_id"

    return ds


def load_tilts(filename, calendar=None, nans=1e25):
    output = list()
    track_id = list()
    times = list()
    with open(filename, "rt") as f:
        # First line
        header = _parse(tilt_header_fmt, f.readline().strip()).named
        ntracks = header["ntracks"]
        nfields = header["nfields"]

        line_fmt = "LEVELS " + " ".join(["{:f}"] * nfields)
        levels = np.asarray(_parse(line_fmt, f.readline().strip()).fixed)

        # Read in each track as an xarray dataset with time as the coordinate
        for n in range(ntracks):
            # Read individual track header
            track_info = _parse(tilt_track_header_fmt, f.readline().strip()).named
            npoints = track_info["npoints"]

            # Add track ID as a variable along the record dimension so that it can be
            # used for groupby
            track_id.extend([track_info["track_id"]] * npoints)

            # Populate time and data line by line
            for m in range(npoints):
                line = f.readline().strip()
                times.append(parse_date(line.split(" ")[0], calendar=calendar))
                output.append(line)

    output = np.genfromtxt(output)
    data_vars = dict(
        tilt=(["record", "levels"], output[:, 1:]),
        track_id=("record", track_id),
        time=("record", times),
    )
    coords = dict(pressure=("levels", levels))

    output = xr.Dataset(data_vars=data_vars, coords=coords)
    output.track_id.attrs["cf_role"] = "trajectory_id"

    if nans is not None:
        output["tilt"] = (
            ["record", "levels"],
            np.where(output.tilt == nans, np.nan, output.tilt),
        )
    output.track_id.attrs["cf_role"] = "trajectory_id"

    return output



"""
Module containing functions to compute track statistics
"""

import numpy as np
import pandas as pd


def track_duration(time, track_ids):
    """
    Compute the duration of each track

    Parameters
    ----------
    time : xarray.DataArray
    track_ids : array_like

    Returns
    -------
    xarray.DataArray
        Duration of each track (in hours)

    """
    duration = (
        time.groupby(track_ids).map(lambda x: x.max() - x.min()).rename("duration")
    )
    duration = duration / np.timedelta64(1, "h")
    duration.attrs["units"] = "hours"
    return duration


def gen_vals(tracks, time, track_id):
    """
    Shows the attributes for the genesis point of each track

    Parameters
    ----------
    tracks : xarray.Dataset
    time : array_like
    track_id : xarray.DataArray

    Returns
    -------
    xarray.Dataset
        Dataset containing only genesis points, with track_id as index.

    """
    # It is 470 times much faster to switch to a dataframe...
    # Use the sortby/groupby with pandas to find the relevant indices in the original
    # Dataset by passing an index (named idx to not clash with "index")
    df = pd.DataFrame(
        data=dict(
            idx=np.arange(len(track_id)),
            time=np.asarray(time),
            track_id=np.asarray(track_id),
        )
    )
    idx = np.asarray(df.sort_values("time").groupby("track_id").first().idx)

    # Could check that track_id is 1d, but the function would already have failed by now
    # if not
    dim = track_id.dims[0]
    tracks = tracks.isel(**{dim: idx})

    # Promote track_id to a coordinate and remove record
    return tracks.assign_coords(**{track_id.name: tracks[track_id.name]}).swap_dims(
        **{dim: track_id.name}
    )


def apex_vals(tracks, variable, track_id, stat="max"):
    """
    Shows the attribute for the extremum point of each track

    Parameters
    ----------
    tracks : xarray.DataSet
    variable : array_like
        The extremum variable
    track_id : xarray.DataArray
    stat : str, optional
        Type of extremum. Can be "min" or "max". The default is "max".

    Raises
    ------
    NotImplementedError
        If another value than "min" and "max" is given to stat

    Returns
    -------
    xarray.Dataset
        Dataset containing only extremum points, with track_id as index.

    """

    # tracks will be sorted along var and then the first line of each track_id will be
    # used
    # asc determines whether the sorting must be ascending (True) or descending (False)
    if stat == "max":
        asc = False
    elif stat == "min":
        asc = True
    else:
        raise NotImplementedError("stat not recognized. Please use one of {min, max}")

    # It is 350 times much faster to switch to a dataframe.
    # Use the same trick as with gen_vals
    df = pd.DataFrame(
        data=dict(
            idx=np.arange(len(variable)),
            var=np.asarray(variable),
            track_id=np.asarray(track_id),
        )
    )
    idx = np.asarray(
        df.sort_values("var", ascending=asc).groupby("track_id").first().idx
    )

    dim = track_id.dims[0]
    tracks = tracks.isel(**{dim: idx})

    # Promote track_id to a coordinate and remove record
    return tracks.assign_coords(**{track_id.name: tracks[track_id.name]}).swap_dims(
        **{dim: track_id.name}
    )



import pytest
import numpy as np

import huracanpy


@pytest.mark.parametrize(
    "filename, kwargs, nvars, ncoords, npoints, ntracks",
    [
        (huracanpy.example_TRACK_file, dict(source="TRACK"), 35, 0, 46, 2),
        (huracanpy.example_TRACK_tilt_file, dict(source="TRACK.tilt"), 3, 1, 46, 2),
        (huracanpy.example_csv_file, dict(), 9, 0, 99, 3),
        (huracanpy.example_parquet_file, dict(), 9, 0, 99, 3),
        (huracanpy.example_TRACK_netcdf_file, dict(), 20, 17, 4580, 86),
        (huracanpy.example_TRACK_netcdf_file, dict(source="netcdf"), 20, 17, 4580, 86),
        (huracanpy.example_TRACK_timestep_file, dict(source="TRACK"), 38, 0, 416, 19),
        (
            huracanpy.example_TRACK_timestep_file,
            dict(source="TRACK", track_calendar=("1940-01-01", 6)),
            38,
            0,
            416,
            19,
        ),
        (
            huracanpy.example_TRACK_timestep_file,
            dict(
                source="TRACK",
                track_calendar=(
                    np.datetime64("1940-01-01"),
                    np.timedelta64(6 * 60 * 60, "s"),
                ),
            ),
            38,
            0,
            416,
            19,
        ),
        (huracanpy.example_TE_file, dict(source="tempestextremes"), 8, 0, 210, 8),
        (
            huracanpy.example_TE_file,
            dict(source="tempestextremes", variable_names=["slp", "wind"]),
            8,
            0,
            210,
            8,
        ),
        (huracanpy.example_CHAZ_file, dict(), 11, 0, 1078, 20),
        (huracanpy.example_MIT_file, dict(), 10, 1, 2138, 11),
        (huracanpy.example_WiTRACK_file, dict(source="witrack"), 14, 0, 3194, 268),
        (None, dict(source="ibtracs", ibtracs_subset="wmo"), 8, 0, 143287, 4540),
        (None, dict(source="ibtracs", ibtracs_subset="usa"), 10, 0, 121806, 4170),
        (huracanpy.example_old_HURDAT_file, dict(source="ecmwf"), 8, 0, 183, 29),
        (
            huracanpy.example_STORM_file,
            dict(
                source="csv",
                names=[
                    "year",
                    "month",
                    "time",
                    "TC_num",
                    "timeStep",
                    "basinID",
                    "lat",
                    "lon",
                    "minP",
                    "Vmax",
                    "Rmax",
                    "cat",
                    "landfall",
                    "dist_land",
                ],
                infer_track_id=["year", "tc_num"],
            ),
            15,
            0,
            3909,
            134,
        ),
        (huracanpy.example_IRIS_file, dict(source="iris"), 10, 0, 40, 2),
        # Two csv file that should load with the same options
        (
            [huracanpy.example_csv_file, huracanpy.example_year_file],
            dict(),
            13,
            0,
            2373,
            92,
        ),
    ],
)
def test_load(filename, kwargs, nvars, ncoords, npoints, ntracks):
    data = _load_with_checked_warnings(filename, **kwargs)

    assert len(data) == nvars
    assert len(data.coords) == ncoords
    assert len(data.time) == npoints
    assert len(data.groupby("track_id")) == ntracks
    assert "record" not in data.coords

    if filename != huracanpy.example_TRACK_tilt_file:
        for name in ["track_id", "time", "lon", "lat"]:
            assert name in data


def _fake_ibtracs_data(url, filename):  # noqa ARG001
    return huracanpy.example_csv_file, None


def test_load_ibtracs_online(monkeypatch):
    with monkeypatch.context() as m:
        from huracanpy._data import ibtracs

        m.setattr(ibtracs, "urlretrieve", _fake_ibtracs_data)
        tracks = huracanpy.load(source="ibtracs", ibtracs_subset="last3years")

    assert len(tracks) == 9
    assert len(tracks.coords) == 0
    # IBTrACS online load skips the second line (first line after header) so this is one
    # less than expected
    assert len(tracks.time) == 98
    assert len(tracks.groupby("track_id")) == 3
    assert "record" not in tracks.coords

    for name in ["track_id", "time", "lon", "lat"]:
        assert name in tracks


@pytest.mark.parametrize(
    "filename, kwargs, error, message",
    [
        ("", dict(source="nonsense"), ValueError, "Source nonsense unsupported"),
        ("", dict(), ValueError, "Source is set to None"),
    ],
)
def test_load_fails(filename, kwargs, error, message):
    with pytest.raises(error, match=message):
        huracanpy.load(filename, **kwargs)


def test_load_rename():
    tracks = huracanpy.load(
        huracanpy.example_csv_file,
        rename=dict(slp="pressure", not_a_variable="should_be_ignored"),
    )

    assert "pressure" in tracks
    assert "slp" not in tracks
    assert "should_be_ignored" not in tracks


def test_load_units():
    tracks = huracanpy.load(huracanpy.example_csv_file, units=dict(slp="Pa"))

    assert tracks.slp.attrs["units"] == "Pa"

    slp_hpa = tracks.slp.metpy.convert_units("hPa")

    np.testing.assert_allclose(tracks.slp.values, slp_hpa.data.magnitude * 100)


def test_load_baselon():
    tracks = huracanpy.load(huracanpy.example_csv_file, baselon=1000)

    assert tracks.lon.max() <= 1360
    assert tracks.lon.min() >= 1000


@pytest.mark.parametrize(
    "filename, source",
    [
        (huracanpy.example_TRACK_file, "TRACK"),
        (huracanpy.example_TRACK_tilt_file, "TRACK.tilt"),
        (huracanpy.example_TRACK_netcdf_file, None),
        (huracanpy.example_csv_file, None),
        (huracanpy.example_parquet_file, None),
        (huracanpy.example_TE_file, "tempestextremes"),
        (huracanpy.example_CHAZ_file, None),
        (huracanpy.example_MIT_file, None),
        (huracanpy.example_WiTRACK_file, "witrack"),
        (None, "ibtracs"),
    ],
)
@pytest.mark.parametrize("extension", ["csv", "nc"])
@pytest.mark.parametrize(
    "muddle, use_accessor", [(False, False), (True, False), (False, True)]
)
def test_save(filename, source, extension, muddle, use_accessor, tmp_path):
    if extension == "csv" and (
        filename == huracanpy.example_TRACK_tilt_file
        or (filename is not None and filename.split(".")[-1] == "nc")
    ):
        pytest.skip(
            "The netCDF file has multiple dimensions so fails because converting to a"
            " dataframe leads to having rows equal to the product of the dimensions"
            " even though the dimensions cover different variables"
        )
    data = _load_with_checked_warnings(filename, source=source)

    # Check that save/load gives the same result when the track_id is not monotonic
    # Caused an issue because they got sorted before
    if muddle:
        data = data.sortby("track_id", ascending=False)

    # Copy the data because save modifies the dataset at the moment
    data_orig = data.copy()

    filename = str(tmp_path / f"tmp_file.{extension}")
    if use_accessor:
        data.hrcn.save(filename)
    else:
        huracanpy.save(data, filename)

    # Check that the original data is not modified by the save function
    _assert_dataset_identical(data_orig, data)

    # Reload the data and check it is still the same
    # Saving as netcdf does force sorting by track_id so apply this
    if extension == "nc":
        data = data.sortby("track_id")
    data_reload = huracanpy.load(filename)
    _assert_dataset_identical(data, data_reload)


def test_save_fails(tracks_csv):
    with pytest.raises(NotImplementedError, match="File format not recognized"):
        huracanpy.save(tracks_csv, "filename.unsupported_extension")


def _load_with_checked_warnings(filename, **kwargs):
    if filename is None:
        if "ibtracs_subset" not in kwargs or kwargs["ibtracs_subset"] == "wmo":
            with (
                pytest.warns(
                    UserWarning,
                    match="This offline function loads a light version of IBTrACS",
                ),
                pytest.warns(
                    UserWarning, match="You are loading the IBTrACS-WMO subset"
                ),
            ):
                data = huracanpy.load(filename, **kwargs)
        else:
            with pytest.warns(
                UserWarning,
                match="This offline function loads a light version of IBTrACS",
            ):
                data = huracanpy.load(filename, **kwargs)
    else:
        data = huracanpy.load(filename, **kwargs)

    return data


def _assert_dataset_identical(ds1, ds2):
    assert len(ds1.variables) == len(ds2.variables)
    assert len(ds1.coords) == len(ds2.coords)
    for var in list(ds1.variables) + list(ds1.coords):
        # Work around for xarray inconsistent loading the data as float or double
        # depending on fill_value and scale_factor
        # np.testing.assert_allclose doesn't work for datetime64, object, or string
        if np.issubdtype(ds1[var].dtype, np.number):
            if ds1[var].dtype != ds2[var].dtype:
                rtol = 1e-6
            else:
                rtol = 0
            np.testing.assert_allclose(
                ds1[var].data.astype(ds2[var].dtype), ds2[var].data, rtol=rtol
            )
        else:
            assert (ds1[var].data == ds2[var].data).all()

    assert len(ds1.attrs) == len(ds2.attrs)
    for attr in ds1.attrs:
        assert ds1.attrs[attr] == ds2.attrs[attr]



import pathlib

import pytest
import xarray as xr

import huracanpy


data_path = pathlib.Path(__file__).parent / "data_interpolate"


@pytest.mark.parametrize("prog_bar", [(True,), (False,)])
def test_interpolate_time(tracks_csv, prog_bar):
    result = huracanpy.interp_time(tracks_csv, tracks_csv.track_id, prog_bar=prog_bar)

    expected = huracanpy.load(str(data_path / "interpolate_time_result.nc"))

    # cf_role="trajectory_id" automatically added when result was saved
    del expected.track_id.attrs["cf_role"]

    xr.testing.assert_allclose(result, expected)



import numpy as np
import xarray as xr


def concat_tracks(
    objs, track_id="track_id", *, prefix=None, start=0, keep_track_id=False, **kwargs
):
    """Concatenate the tracks while making sure the track_id remains unique for each
    track

    Parameters
    ----------
    objs : sequence of xarray.Dataset
        The tracks to concatenate
    track_id : str, default="track_id"
        The name of the track_id variable in the tracks to concatenate
    prefix : str, optional
        A string to start each track_id to make them unique. Needs to be a
        python-formattable string e.g. `"{}_"` will result in track_id's that look like
        `{n}_{track_id}`, where `n` is the count of the DataArray in objs and `track_id`
        is the original track_id.
    start : int, optional
        The first entry for the new track IDs. If prefix is None, then the track_ids
        will have values from `start` to `start` + number_of_tracks
    keep_track_id : bool or str, default=False
        Save the original track_id as a new variable in the output Dataset. If `str`
        this gives the variable name, "track_id_original" otherwise
    **kwargs :
        Passed to :py:func:`xarray.concat`

    Returns
    -------

    """
    all_tracks = []

    current_track_id = start
    for n, tracks in enumerate(objs, start=start):
        track_id_old = tracks[track_id]

        if prefix is not None:
            prefix_ = xr.DataArray(prefix.format(n) + "{}")
            tracks = tracks.assign(
                {track_id: prefix_.str.format(track_id_old.astype(str))}
            )

        else:
            tracks = _reset_track_id(
                tracks,
                track_id_old,
                start=current_track_id,
                keep_original=keep_track_id,
            )
            current_track_id = tracks[track_id].values.max() + 1

        all_tracks.append(tracks)

    return xr.concat(all_tracks, dim=track_id_old.dims[0], **kwargs)


def _reset_track_id(tracks, track_ids, start=0, *, keep_original=False):
    """Replace the track IDs with an ascending sequence of numbers

    Examples
    >>> tracks1 = _reset_track_id(tracks1,tracks1.track_id)
    >>> tracks2 = _reset_track_id(
    >>>     tracks2,tracks2.track_id,start=tracks1.track_id.values.max() + 1
    >>> )
    >>> all_tracks = xr.concat([tracks1, tracks2], dim="record")

    Parameters
    ----------
    tracks : xarray.Dataset
    track_ids : array_like
        The current track IDs
    start : int, optional
        The first entry for the new track IDs. The track_ids will have values from
        `start` to `start` + number_of_tracks
    keep_original : bool or str, default=False
        Save the original track_id as a new variable in the output Dataset. If `str`
        this gives the variable name, "track_id_original" otherwise

    Returns
    -------
    xarray.Dataset:
        A copy of the input dataset with reassigned track_ids
    """
    if track_ids.ndim != 1:
        raise ValueError("track_ids must be 1d")

    _, new_track_ids = np.unique(track_ids, return_inverse=True)

    # Default name for original track id if not specified
    if keep_original is True:
        keep_original = "track_id_original"

    if isinstance(keep_original, str):
        tracks = tracks.assign(**{keep_original: (track_ids.dims[0], track_ids.values)})

    tracks = tracks.assign(
        **{track_ids.name: (track_ids.dims[0], start + new_track_ids)}
    )

    return tracks



"""
Module with function to compute TC-specific categories
"""

import warnings

from metpy.units import units
from metpy.xarray import preprocess_and_wrap

from ..info import category
from .._metpy import dequantify_results, validate_units
from ._conventions import _thresholds


def saffir_simpson_category(wind, convention="Saffir-Simpson", wind_units="m s-1"):
    """
    Function to determine the Saffir-Simpson Hurricane Scale (SSHS) category.

    Parameters
    ----------
    wind : array_like
        10-minutes averaged 10m wind in m/s

    convention : str

    wind_units : str, default="m s-1"
        The units of the input array if they are not already provided by the attributes

    Returns
    -------
    array_like
        The category series.
        You can append it to your tracks by running
        tracks["sshs"] = get_sshs_cat(tracks.wind)
    """
    return category(
        wind,
        bins=_thresholds[convention]["bins"],
        labels=_thresholds[convention]["labels"],
        variable_units=wind_units,
    )


@dequantify_results
@preprocess_and_wrap(wrap_like="slp")
def pressure_category(slp, convention="Klotzbach", slp_units=None):
    """
    Determine the pressure category according to selected convention.

    Parameters
    ----------
    slp : xarray.DataArray
        Minimum Sea-level Pressure series in hPa
    convention : str
        Name of the classification convention you want to use.
            * Klotzbach (default)
            * Simpson
    slp_units : str, default="hPa"
        The units of the input array if they are not already provided by the attributes

    Returns
    -------
    xarray.DataArray
        The category series.
        You can append it to your tracks by running
        tracks["cat"] = get_pressure_cat(tracks.slp)

    """
    # Don't automatically switch units to Pa if they have been explicitly set to Pa,
    # even if they seem wrong
    if (not isinstance(slp, units.Quantity) or slp.unitless) and slp_units is None:
        if slp.magnitude.min() > 10000:
            warnings.warn(
                "Caution, pressure are likely in Pa, they are being converted to hPa "
                "for categorization. In future specify the units explicitly by passing "
                'slp_units="Pa" to this function or setting '
                'slp.attrs["units"] = "Pa"'
            )
            slp_units = "Pa"
        else:
            slp_units = "hPa"

    slp = validate_units(slp, expected_units=slp_units)

    return category(
        slp,
        bins=_thresholds[convention]["bins"],
        labels=_thresholds[convention]["labels"],
        variable_units=slp_units,
    )



import numpy as np
import xarray as xr


# http://cfconventions.org/Data/cf-conventions/cf-conventions-1.11/cf-conventions.html#_contiguous_ragged_array_representation_of_trajectories
# The CF convention for saving the data is that the track_id or equivalent variable has
# length of number of tracks (trajectories) and is labelled with cf_role="trajectory"
# and a rowSize variable with the attribute sample_dimension="obs" (or equivalent of
# "obs") to tell us how to split apart the tracks.
# For working on track data we extend the track_id to the full length of all the
# trajectories to allow us to use groupby() and sel() with track_id to work by
# individual tracks. So we need to replace the "track_id" or equivalent variable when
# loading or saving the data
def load(filename, rename, **kwargs):
    dataset = xr.open_dataset(filename, **kwargs)

    # xarray.Dataset.rename only accepts keys that are actually in the dataset
    rename = {
        key: rename[key] for key in rename if key in dataset or key in dataset.dims
    }
    if len(rename) > 0:
        dataset = dataset.rename(rename)

    # Assume ragged array. If the CF trajectory_id and sample dimension can't be found
    # try reshaping from a 2d array
    try:
        dataset = stretch_trid(dataset)
    except ValueError:
        dataset = as1d(dataset)

    return dataset


def save(dataset, filename):
    # Find the variable with cf_role=trajectory_id
    trajectory_id = _find_trajectory_id(dataset)

    # Get the name of the sample dimension. The name "record" has been used in the load
    # functions, but we don't need to assume that is the name. It may be different when
    # loaded from other netCDF files
    sample_dimension = trajectory_id.dims

    if len(sample_dimension) == 1:
        sample_dimension = sample_dimension[0]
    else:
        raise ValueError(
            f"{trajectory_id.name} spans multiple dimensions, should be 1d"
        )

    # Sort by trajectory_id so each track can be described by the first index and
    # number of elements of the unique trajectory id
    dataset = dataset.sortby(trajectory_id.name)
    trajectory_ids = np.unique(trajectory_id)
    rowsize = [np.count_nonzero(trajectory_id == x) for x in trajectory_ids]

    dataset[trajectory_id.name] = ("trajectory", trajectory_ids)
    dataset[trajectory_id.name].attrs = trajectory_id.attrs
    dataset["rowSize"] = ("trajectory", rowsize)
    dataset["rowSize"].attrs["sample_dimension"] = sample_dimension

    dataset.to_netcdf(filename)


def stretch_trid(dataset):
    # Find the trajectory_id and rowSize variables from their CF labels
    trajectory_id = _find_trajectory_id(dataset)
    rowsize = _find_rowsize(dataset)
    sample_dimension = rowsize.attrs["sample_dimension"]

    # Stretch the trajectory_id out along the sample dimension
    trajectory_id_stretched = []
    for npoints, tr_id in zip(rowsize.data, trajectory_id.data):
        trajectory_id_stretched.extend([tr_id] * npoints)

    dataset = dataset.drop_vars([trajectory_id.name, rowsize.name])

    dataset[trajectory_id.name] = (sample_dimension, trajectory_id_stretched)
    # Keep attributes (add cf_role if not already there)
    dataset[trajectory_id.name].attrs = trajectory_id.attrs
    dataset[trajectory_id.name].attrs["cf_role"] = "trajectory_id"

    return dataset


def as1d(dataset):
    # Stack 2d dimensions into a record dimension
    dims = dataset.lon.dims
    dataset = dataset.stack(record=dims)

    # Record only as an index, not a coordinate
    record = dataset.record
    dataset = dataset.drop_vars(["record", *dims])

    # Set old dims as variables
    for dim in dims:
        # If it is a dimension, add it as a variable
        dataset[dim] = ("record", record[dim].values)

    # Remove data that is only nans
    dataset = dataset.where(~np.isnan(dataset.lon), drop=True)

    # Add cf role to track_id
    dataset.track_id.attrs["cf_role"] = "trajectory_id"

    return dataset.sortby(["track_id", "time"])


def _find_trajectory_id(dataset):
    # Find the variable with cf_role=trajectory_id
    trajectory_id = [
        dataset[var]
        for var in dataset.variables
        if "cf_role" in dataset[var].attrs
        and dataset[var].attrs["cf_role"] == "trajectory_id"
    ]

    if len(trajectory_id) == 1:
        return trajectory_id[0]
    else:
        if "track_id" in dataset:
            return dataset["track_id"]
        else:
            raise ValueError(
                f"Found {len(trajectory_id)} variables with cf_role=trajectory_id."
                f"Should be exactly one."
            )


def _find_rowsize(dataset):
    # Find the variable with sample_dimension="obs" (or equivalent)
    rowsize = [
        dataset[var]
        for var in dataset.variables
        if "sample_dimension" in dataset[var].attrs
    ]

    if len(rowsize) == 1:
        return rowsize[0]
    else:
        raise ValueError(
            f"Found {len(rowsize)} variables to map to row size. Should be exactly one"
        )



"""
Module containing functions to compute lifecycle stage
"""

import pandas as pd


def time_from_genesis(time, track_ids):
    """
    Output the time since genesis for each TC point.

    Parameters
    ----------
    time: xr.DataArray
    track_ids: xr.DataArray

    Returns
    -------
    xarray.DataArray
        The time_from_genesis series.
        You can append it to your tracks by running

        >>> tracks["time_from_genesis"] = time_from_genesis(
        >>>     tracks.time, tracks.track_id
        >>> )

    """
    data_df = pd.DataFrame({"time": time, "track_id": track_ids})
    data_df = data_df.merge(
        data_df.groupby("track_id").time.min(),
        on="track_id",
        suffixes=["_actual", "_gen"],
    )
    time_from_start = data_df.time_actual - data_df.time_gen
    return (
        time_from_start.to_xarray()
        .rename({"index": track_ids.dims[0]})
        .drop_vars(track_ids.dims[0])
        .rename("time_from_genesis")
    )


def time_from_apex(time, track_ids, intensity_var, stat="max"):
    """The time relative to a maxima/minima in a given variable for each individual
    track

    >>> time_from_apex(tracks.time, tracks.track_id, tracks.wind, stat="max")

    Parameters
    ----------
    time : array_like
    track_ids : xarray.DataArray
    intensity_var : array_like
    stat : str, optional
        Take either the maxima ("max") or minima ("min") of `intensity_var`. Default is
        "max"

    Returns
    -------
    xarray.DataArray

    """
    if stat == "max":
        asc = False
    elif stat == "min":
        asc = True
    else:
        raise NotImplementedError("stat not recognized. Please use one of {min, max}")
    data_df = pd.DataFrame(
        {"time": time, "track_id": track_ids, "intensity": intensity_var}
    )
    extr = data_df.sort_values("intensity", ascending=asc).groupby("track_id").first()
    data_df = data_df.merge(extr, on="track_id", suffixes=["_actual", "_extr"])
    time_from_extr = data_df.time_actual - data_df.time_extr
    return (
        time_from_extr.to_xarray()
        .rename({"index": track_ids.dims[0]})
        .drop_vars(track_ids.dims[0])
        .rename("time_from_extremum")
    )



"""Huracanpy module for tracks diagnostics"""

__all__ = [
    "density",
    "track_duration",
    "gen_vals",
    "apex_vals",
    "time_from_genesis",
    "time_from_apex",
    "delta",
    "rate",
    "distance",
    "translation_speed",
    "azimuth",
]

from ._density import density
from ._track_stats import (
    track_duration,
    gen_vals,
    apex_vals,
)
from ._lifecycle import time_from_genesis, time_from_apex
from ._rates import delta, rate
from ._translation import distance, translation_speed, azimuth



"""Tropical Cyclone specific calculations"""

__all__ = [
    "ace",
    "pace",
    "saffir_simpson_category",
    "pressure_category",
    "beta_drift",
]

from ._ace import ace, pace
from ._category import saffir_simpson_category, pressure_category
from ._beta_drift import beta_drift



import numpy as np
import xarray as xr

__all__ = ["trackswhere", "sel_id"]


def sel_id(tracks, track_ids, track_id):
    """Select a track or tracks from a dataset of tracks by ID

    Parameters
    ----------
    tracks : xarray.Dataset
    track_ids : xarray.DataArray
        The track_ids corresponding to the tracks Dataset
    track_id : Any
        The track ID or IDs to match in track_ids.

    Returns
    -------
    xarray.Dataset

    """
    if track_ids.ndim != 1:
        raise ValueError("track_ids must be 1d")

    if np.isscalar(track_id):
        track_id = [track_id]

    dim = track_ids.dims[0]
    idx = np.where(np.isin(track_ids, track_id))[0]

    return tracks.isel(**{dim: idx})


def trackswhere(tracks, track_ids, condition):
    """Subset tracks that verify a condition.

    e.g select all tracks that are solely in the Northern hemisphere:

    >>> tracks_subset = huracanpy.trackswhere(
    >>>     tracks, tracks.track_id, lambda x: (x.lat > 0).all()
    >>> )

    e.g. select all tracks that are category 2 at least once in their lifetime:

    >>> track_subset = huracanpy.trackswhere(
        tracks, tracks.track_id, lambda track: track.pressure_category.max() >= 2
        )

    Parameters
    ----------
    tracks : xarray.Dataset
    track_ids : xarray.DataArray
        The track_ids corresponding to the tracks Dataset
    condition : callable
        A function that takes an `xarray.Dataset` of an individual track and returns
        True or False

    Returns
    -------
    xarray.Dataset
        A dataset with the subset of tracks from the input that match the given criteria

    """
    if track_ids.ndim != 1:
        raise ValueError("track_ids must be 1d")

    track_groups = [
        track for track_id, track in tracks.groupby(track_ids) if condition(track)
    ]

    return xr.concat(track_groups, dim=track_ids.dims[0])



from inspect import getmembers, isfunction

import matplotlib.pyplot as plt
from matplotlib.testing.decorators import check_figures_equal
import numpy as np
import pytest
import xarray as xr

import huracanpy


# Functions not in the accessor
_intentionally_missing = [
    "load",
    "concat_tracks",
    # add_ functions that would have output with a different shape
    "add_apex_vals",
    "add_gen_vals",
    "add_density",
    "add_track_duration",
    # plot_ functions that are for multiple datasets
    "plot_doughnut",
    "plot_venn",
]


def _get_function_args(tracks, function_args):
    # Get the function arguments as arrays. Use "all" as a wildcard for the full dataset
    # Keep other types of arguments as is
    return [
        var if not isinstance(var, str) else tracks if var == "all" else tracks[var]
        for var in function_args
    ]


# %% DataArrayAccessor
def test_nunique():
    data = huracanpy.load(huracanpy.example_csv_file)

    n_tracks = data.track_id.hrcn.nunique()
    assert n_tracks == 3


@pytest.mark.parametrize("call_type", ["get", "add"])
@pytest.mark.parametrize(
    "function, function_args, accessor_function_kwargs",
    [
        (huracanpy.info.hemisphere, ["lat"], {}),
        (huracanpy.info.basin, ["lon", "lat"], {}),
        (huracanpy.info.is_land, ["lon", "lat"], {}),
        (huracanpy.info.is_ocean, ["lon", "lat"], {}),
        (huracanpy.info.country, ["lon", "lat"], {}),
        (huracanpy.info.continent, ["lon", "lat"], {}),
        (
            huracanpy.info.category,
            ["wind10", [0, 10, 20, 30], [0, 1, 2]],
            {"var_name": "wind10", "bins": [0, 10, 20, 30], "labels": [0, 1, 2]},
        ),
        (huracanpy.info.season, ["track_id", "lat", "time"], {}),
        (huracanpy.info.time_components, ["time"], {}),
        (huracanpy.calc.density, ["lon", "lat"], {}),
        (huracanpy.calc.track_duration, ["time", "track_id"], {}),
        (huracanpy.calc.gen_vals, ["all", "time", "track_id"], {}),
        (
            huracanpy.calc.apex_vals,
            ["all", "wind10", "track_id"],
            {"var_name": "wind10"},
        ),
        (huracanpy.calc.time_from_genesis, ["time", "track_id"], {}),
        (
            huracanpy.calc.time_from_apex,
            ["time", "track_id", "wind10"],
            {"intensity_var_name": "wind10"},
        ),
        (huracanpy.calc.delta, ["wind10", "track_id"], {"var_name": "wind10"}),
        (
            huracanpy.calc.delta,
            ["wind10"],
            {"var_name": "wind10", "track_id_name": None},
        ),
        (huracanpy.calc.rate, ["wind10", "time", "track_id"], {"var_name": "wind10"}),
        (
            huracanpy.calc.rate,
            ["wind10", "time"],
            {"var_name": "wind10", "track_id_name": None},
        ),
        (huracanpy.calc.distance, ["lon", "lat", "track_id"], {}),
        (huracanpy.calc.distance, ["lon", "lat"], {"track_id_name": None}),
        (huracanpy.calc.azimuth, ["lon", "lat", "track_id"], {}),
        (huracanpy.calc.azimuth, ["lon", "lat"], {"track_id_name": None}),
        (huracanpy.calc.translation_speed, ["lon", "lat", "time", "track_id"], {}),
        (
            huracanpy.calc.translation_speed,
            ["lon", "lat", "time"],
            {"track_id_name": None},
        ),
        (huracanpy.tc.ace, ["wind10"], {"wind_name": "wind10"}),
        (
            huracanpy.tc.ace,
            ["wind10", "track_id"],
            {"wind_name": "wind10", "sum_by": "track_id"},
        ),
        (
            huracanpy.tc.pace,
            ["slp", "wind10"],
            {"pressure_name": "slp", "wind_name": "wind10"},
        ),
        (huracanpy.tc.saffir_simpson_category, ["wind10"], {"wind_name": "wind10"}),
        (
            huracanpy.tc.pressure_category,
            ["slp"],
            {"slp_name": "slp", "slp_units": "Pa"},
        ),
        # SLP for RMW is nonsense, but I'm just testing that the results are the same
        # and SLP is the best variable available for a reasonable order of magnitude
        (
            huracanpy.tc.beta_drift,
            ["lat", "wind10", "slp"],
            {"wind_name": "wind10", "rmw_name": "slp"},
        ),
    ],
)
def test_accessor_methods_match_functions(
    tracks_csv,
    function,
    function_args,
    accessor_function_kwargs,
    call_type,
):
    accessor_name = f"{call_type}_{function.__name__}"
    # Skip functions that only have a "get_" version
    if accessor_name in _intentionally_missing:
        pytest.skip(f"Accessor function {accessor_name} does not exist")
    if accessor_name == "add_ace" and "sum_by" in accessor_function_kwargs:
        pytest.skip(f"sum_by not a valid argument for {accessor_name}")

    # Call the huracanpy function
    function_args = _get_function_args(tracks_csv, function_args)
    if function == huracanpy.tc.pressure_category:
        with pytest.warns(UserWarning, match="Caution, pressure are likely in Pa"):
            result = function(*function_args)
    else:
        if (
            "track_id_name" in accessor_function_kwargs
            and accessor_function_kwargs["track_id_name"] is None
        ):
            with pytest.warns(UserWarning, match="track_id is not provided"):
                result = function(*function_args)
        else:
            result = function(*function_args)

    # Call the accessor method
    if (
        "track_id_name" in accessor_function_kwargs
        and accessor_function_kwargs["track_id_name"] is None
    ):
        with pytest.warns(UserWarning, match="track_id is not provided"):
            result_accessor = getattr(tracks_csv.hrcn, accessor_name)(
                **accessor_function_kwargs
            )
    else:
        result_accessor = getattr(tracks_csv.hrcn, accessor_name)(
            **accessor_function_kwargs
        )

    # Special case for PACE returning the values and the model
    if function == huracanpy.tc.pace:
        assert result[1] == result_accessor[1]
        result = result[0]
        result_accessor = result_accessor[0]

    # When using the "add_" method a new Dataset is returned with the variable added
    # The naming of the new variable is either simply the function name (minus "add_")
    # or the function name plus the name of the variable specified if it can be applied
    # to different variables
    if call_type == "add":
        # Special cases for multiple variables being added
        if accessor_name == "add_time_components":
            varname = ["year", "month", "day", "hour"]
            result = xr.Dataset({name: result[n] for n, name in enumerate(varname)})
        elif accessor_name == "add_beta_drift":
            varname = ["v_drift", "theta_drift"]
            result = xr.Dataset({name: result[n] for n, name in enumerate(varname)})
        else:
            varname = function.__name__
        if "var_name" in accessor_function_kwargs:
            varname = f"{varname}_{accessor_function_kwargs['var_name']}"
        result_accessor = result_accessor[varname]

    # Check that the function and method return identical results
    assert type(result) is type(result_accessor), (
        "accessor return type differs from function"
    )
    if isinstance(result, xr.Dataset):
        xr.testing.assert_identical(result, result_accessor)
    else:
        np.testing.assert_equal(
            np.asarray(result),
            np.asarray(result_accessor),
            err_msg="accessor output differs from function output",
        )


@check_figures_equal()
@pytest.mark.parametrize(
    "function, function_args, accessor_function_kwargs",
    [
        (huracanpy.plot.density, [], {}),
        (huracanpy.plot.fancyline, ["lon", "lat", "wind10"], {"colors": "wind10"}),
        (
            huracanpy.plot.tracks,
            ["lon", "lat", "wind10"],
            {"intensity_var_name": "wind10"},
        ),
        (huracanpy.plot.tracks, ["lon", "lat"], {}),
    ],
)
def test_accessor_plot_methods_match_functions(
    fig_test, fig_ref, tracks_csv, function, function_args, accessor_function_kwargs
):
    accessor_name = f"plot_{function.__name__}"

    if function is huracanpy.plot.density:
        function_args = [huracanpy.calc.density(tracks_csv.lon, tracks_csv.lat)]
    elif function is huracanpy.plot.fancyline:
        function_args = [
            _get_function_args(track, function_args)
            for track_id, track in tracks_csv.groupby("track_id")
        ]
    else:
        function_args = _get_function_args(tracks_csv, function_args)

    plt.figure(fig_ref)
    ax = plt.gca()

    if function is huracanpy.plot.fancyline:
        [function(*args, ax=ax) for args in function_args]
    elif (
        function is huracanpy.plot.tracks
        and "intensity_var_name" not in accessor_function_kwargs
    ):
        with pytest.warns(
            UserWarning, match="Ignoring `palette` because no `hue` variable"
        ):
            function(*function_args, ax=ax)
    else:
        function(*function_args, ax=ax)

    plt.figure(fig_test)
    ax = plt.gca()

    if (
        function is huracanpy.plot.tracks
        and "intensity_var_name" not in accessor_function_kwargs
    ):
        with pytest.warns(
            UserWarning, match="Ignoring `palette` because no `hue` variable"
        ):
            getattr(tracks_csv.hrcn, accessor_name)(**accessor_function_kwargs, ax=ax)
    else:
        getattr(tracks_csv.hrcn, accessor_name)(**accessor_function_kwargs, ax=ax)


def test_inferred_track_id(tracks_csv):
    track_id = tracks_csv.hrcn.get_inferred_track_id("track_id")
    xr.testing.assert_equal(track_id, tracks_csv.track_id)


def test_interp_methods():
    data = huracanpy.load(huracanpy.example_csv_file)
    interpolated_data_acc = data.hrcn.interp_time(
        freq="1h", track_id_name="track_id", prog_bar=False
    )
    expected_interpolated_data = huracanpy.interp_time(
        data, data.track_id, freq="1h", prog_bar=False
    )
    np.testing.assert_array_equal(
        interpolated_data_acc.time, expected_interpolated_data.time
    )


def test_accessor_sel_id(tracks_csv):
    result = tracks_csv.hrcn.sel_id(0)
    expected = huracanpy.sel_id(tracks_csv, tracks_csv.track_id, 0)

    xr.testing.assert_identical(result, expected)


def test_accessor_trackswhere(tracks_csv):
    result = tracks_csv.hrcn.trackswhere(lambda x: (x.track_id != 1).any())
    expected = huracanpy.trackswhere(
        tracks_csv, tracks_csv.track_id, lambda x: (x.track_id != 1).any()
    )

    xr.testing.assert_identical(result, expected)


def test_accessor_namespace_matches():
    # Functions at the top level have the same name as in the module
    expected_functions = [m[0] for m in getmembers(huracanpy) if isfunction(m[1])]

    # Functions in calc, info, and tc should have get_ and add_ equivalents in the
    # accessor
    expected_functions += [
        f"{prefix}_{m[0]}"
        for prefix in ["get", "add"]
        for module in [huracanpy.calc, huracanpy.info, huracanpy.tc]
        for m in getmembers(module)
        if isfunction(m[1])
    ]

    # Function in plot should be named plot_ in the accessor
    expected_functions += [
        f"plot_{m[0]}" for m in getmembers(huracanpy.plot) if isfunction(m[1])
    ]

    # Remove functions that do not go in the accessor
    for func in _intentionally_missing:
        expected_functions.remove(func)

    # The names of functions actually available on the accessor
    accessor_functions = [
        m[0]
        for m in getmembers(huracanpy._accessor.HuracanPyDatasetAccessor)
        if isfunction(m[1]) and m[0][0] != "_"
    ]

    if sorted(expected_functions) != sorted(accessor_functions):
        missing = [
            func for func in expected_functions if func not in accessor_functions
        ]
        extras = [func for func in accessor_functions if func not in expected_functions]

        raise ValueError(
            "Module and accessor functions do not match\n"
            + "Functions missing from accessor\n    - "
            + "\n    - ".join(missing)
            + "\nExtra functions in the accessor\n    - "
            + "\n    - ".join(extras)
        )



import numpy as np
import pytest

import huracanpy


def test_sel_id(tracks_csv):
    # Check that only the points with the track id are extracted each time and all of
    # the points are extracted once
    npoints = 0
    for track_id in np.unique(tracks_csv.track_id):
        tracks_subset = huracanpy.sel_id(tracks_csv, tracks_csv.track_id, track_id)
        np.testing.assert_array_equal(tracks_subset.track_id, track_id)
        npoints += len(tracks_subset.record)

    assert npoints == len(tracks_csv.record)


def test_sel_id_fails(tracks_csv):
    with pytest.raises(ValueError, match="track_ids must be 1d"):
        huracanpy.sel_id(
            tracks_csv,
            np.asarray([tracks_csv.track_id.values, tracks_csv.track_id.values]),
            0,
        )


def test_sel_id_array(tracks_csv):
    npoints = 0
    for track_id in [[0, 1], [2]]:
        tracks_subset = huracanpy.sel_id(tracks_csv, tracks_csv.track_id, track_id)
        assert np.isin(tracks_subset.track_id, track_id).all()
        npoints += len(tracks_subset.record)

    assert npoints == len(tracks_csv.record)


def test_trackswhere():
    tracks = huracanpy.load(huracanpy.example_csv_file)

    tracks["category"] = huracanpy.tc.pressure_category(tracks.slp, slp_units="Pa")

    tracks_subset = huracanpy.trackswhere(
        tracks, tracks.track_id, lambda track: track.category.max() >= 2
    )

    assert set(tracks_subset.track_id.data) == {0, 2}


def test_trackswhere_fails(tracks_csv):
    with pytest.raises(ValueError, match="track_ids must be 1d"):
        huracanpy.trackswhere(
            tracks_csv,
            np.asarray([tracks_csv.track_id.values, tracks_csv.track_id.values]),
            lambda track: (track.track_id == 0).all(),
        )



"""Matching functions"""

import pandas as pd
import numpy as np
from haversine import haversine_vector, Unit
from itertools import combinations


def match(
    tracksets, names=["1", "2"], max_dist=300, min_overlap=0, tracks1_is_ref=False
):
    """
    Match the provided track sets between them.

    Parameters
    ----------
    tracksets : list[xarray.Dataset]
        list of track datasets to match together. Must be of length two or more.
    names : list, optional
        list of track datasets names. Must be the same size as tracksets. The default is
        ['1','2'].
    max_dist : float, optional
        Threshold for maximum distance between two tracks. The default is 300.
    min_overlap : int, optional
        Minimum number of overlapping time steps for matching. The default is 0.

    Returns
    -------
    pandas.DataFrame
        Dataframe containing the matching tracks with
            the id from corresponding datasets
            the number of matching time steps (if only two datasets provided)
            the distance between two tracks (if only two datasets provided)

    """

    # Check input
    if len(tracksets) < 2:
        raise ValueError("You must provide at least two datasets to match")
    if len(names) != len(tracksets):
        raise ValueError(
            "Number of names provided do not correspond to the number of track sets"
        )

    # Two track sets
    if len(tracksets) == 2:
        return _match_pair(*tracksets, *names, max_dist, min_overlap, tracks1_is_ref)
    # More than two track sets
    else:
        return _match_multiple(tracksets, names, max_dist, min_overlap, tracks1_is_ref)


def _match_pair(
    tracks1,
    tracks2,
    name1="1",
    name2="2",
    max_dist=300,
    min_overlap=0,
    tracks1_is_ref=False,
):
    """

    Parameters
    ----------
    tracks1 (pd.DataFrame): the first track set to match
    tracks2 (pd.DataFrame): the second tracks set to match
    name1 (str): Suffix for the first dataframe
    name2 (str): Suffix for the second dataframe
    max_dist (float) : Threshold for maximum distance between two tracks
    min_overlap (int) : Minimum number of overlapping time steps for matching

    Returns
    -------
    pd.DataFrame
        Dataframe containing the matching tracks with
            the id from both datasets
            the number of matching time steps
            the distance between two tracks
    """

    # Prepare data
    tracks1, tracks2 = (
        tracks1[["track_id", "lon", "lat", "time"]].to_dataframe(),
        tracks2[["track_id", "lon", "lat", "time"]].to_dataframe(),
    )
    tracks1["lon"] = np.where(tracks1.lon > 180, tracks1.lon - 360, tracks1.lon)
    tracks2["lon"] = np.where(tracks2.lon > 180, tracks2.lon - 360, tracks2.lon)

    # Find corresponding points (same time step, less than max_dist km)
    merged = pd.merge(tracks1, tracks2, on="time")

    if len(merged) > 0:  # if there exist matching points, continue
        x = np.concatenate([[merged.lat_x], [merged.lon_x]]).T
        y = np.concatenate([[merged.lat_y], [merged.lon_y]]).T
        merged["dist"] = haversine_vector(x, y, unit=Unit.KILOMETERS)
        merged = merged[merged.dist <= max_dist]
        # Compute temporal overlap
        temp = (
            merged.groupby(["track_id_x", "track_id_y"])[["dist"]]
            .count()
            .rename(columns={"dist": "temp"})
        )
        # Build a table of all pairs of tracks sharing at least one point
        matches = (
            merged[["track_id_x", "track_id_y"]]
            .drop_duplicates()
            .join(temp, on=["track_id_x", "track_id_y"])
        )
        matches = matches[matches.temp >= min_overlap]
        dist = merged.groupby(["track_id_x", "track_id_y"])[["dist"]].mean()
        matches = matches.merge(dist, on=["track_id_x", "track_id_y"])

        # Treat duplicates if required
        if tracks1_is_ref:
            # Treat the duplicates where one tracks2 track has several corresponding
            # tracks1:
            # Keep the couple with the longest overlap
            matches = (
                matches.sort_values("temp", ascending=False)
                .groupby("track_id_y")
                .first()
                .reset_index()
            )

        # Rename columns before output
        matches = matches.rename(
            columns={"track_id_x": "id_" + name1, "track_id_y": "id_" + name2}
        )
        return matches

    else:  # if there exist no matching points, return empty dataframe
        return pd.DataFrame(columns=["id_" + name1, "id_" + name2, "temp", "dist"])


def _match_multiple(
    datasets,
    names,
    max_dist=300,
    min_overlap=0,
    tracks1_is_ref=False,
):
    """
    Function to match any number of tracks sets

    Parameters
    ----------
    datasets : list of xr.Dataset
        list of the sets to be matched.
    names : list of str
        labels for the datasets. names must have the same length as datasets
    max_dist : float
        Threshold for maximum distance between two tracks
    min_overlap : int
        Minimum number of overlapping time steps for matching

    Raises
    ------
    NotImplementedError
        If two datasets have no match.

    Returns
    -------
    M : pd.dataframe
        table of matching tracks among all the datasets

    """
    matches = pd.DataFrame(columns=["id_" + n for n in names[:2]])
    for names_pair, dataset_pair in zip(
        combinations(names, 2), combinations(datasets, 2)
    ):
        m = _match_pair(
            *dataset_pair,
            *names_pair,
            max_dist,
            min_overlap,
            tracks1_is_ref=tracks1_is_ref * (names_pair[0] == names[0]),
        )
        if len(m) == 0:
            raise NotImplementedError(
                "For the moment, the case where two datasets have no match is not"
                "handled. Problem raised by datasets " + str(names_pair)  # TODO
            )
        matches = matches.merge(
            m[["id_" + names_pair[0], "id_" + names_pair[1]]], how="outer"
        )
    return matches


# TODO: Deal with duplicates: merge, max...?



"""Venn diagrams for tracks matching visualisation"""

from matplotlib_venn import venn2, venn2_circles, venn3, venn3_circles
import numpy as np


def venn(datasets, match, labels, colors=None, circle_color="k"):
    """
    Plot venn diagram to compare the datasets.

    Parameters
    ----------
    datasets : list of xr.dataset
        list of the datasets compared.
    match : pd.DataFrame
        match dataframe issued from match_pair or match_multiple.
    labels : list of str
        labels of the datasets.
    colors : list of str, optional
        list of colors to be used for each dataset. The default is None.
    circle_color : str, optional
        color of the overlaid circles. The default is "k".

    Raises
    ------
    NotImplementedError
        If more than three or less than two datasets are given.

    Returns
    -------
    None.

    """
    if len(datasets) == 2:
        f = _venn_2datasets
    elif len(datasets) == 3:
        f = _venn_3datasets
    else:
        raise NotImplementedError(
            "We cannot plot Venn diagrams for more than 3 datasets."
        )

    if len(datasets) != len(labels):
        raise ValueError("datasets and labels must have the same length")

    if colors is None:
        colors = ["w"] * len(datasets)
    else:
        if len(colors) != len(datasets):
            raise ValueError("datasets and colors must have the same length")
    f(*datasets, match, colors, labels, circle_color)


def _venn_2datasets(data1, data2, match, colors, labels=None, circle_color="k"):
    n1 = len(np.unique(data1.track_id.values))  # Number of tracks in dataset 1
    n2 = len(np.unique(data2.track_id.values))  # Number of tracks in dataset 2
    m = len(match)  # Number of tracks matching
    venn2((n1 - m, n2 - m, m), set_colors=colors, set_labels=labels)
    venn2_circles((n1 - m, n2 - m, m), color=circle_color)


def _venn_3datasets(
    data1, data2, data3, matches, colors, labels=None, circle_color="k"
):
    n1 = len(np.unique(data1.track_id.values))  # Number of tracks in dataset 1
    n2 = len(np.unique(data2.track_id.values))  # Number of tracks in dataset 2
    n3 = len(np.unique(data3.track_id.values))  # Number of tracks in dataset 3

    m_not1 = len(matches[matches.iloc[:, 0].isna()])
    m_not2 = len(matches[matches.iloc[:, 1].isna()])
    m_not3 = len(matches[matches.iloc[:, 2].isna()])
    m_all = len(matches[matches.isna().sum(axis=1) == 0])

    subsets = (
        (n1 - m_all - m_not2 - m_not3),
        (n2 - m_all - m_not1 - m_not3),
        m_not3,
        (n3 - m_all - m_not1 - m_not2),
        m_not2,
        m_not1,
        m_all,
    )

    venn3(
        subsets,
        set_labels=labels,
        set_colors=colors,
    )

    venn3_circles(subsets, color=circle_color)



"""Module for functions related to the ibtracs database"""

from urllib.request import urlretrieve
import warnings
import pathlib

from . import _csv

here = pathlib.Path(__file__).parent
ibdata_dir = here / "_ibtracs_files/"

wmo_file = str(ibdata_dir / "wmo.csv")
jtwc_file = str(ibdata_dir / "jtwc.csv")

online_default_kwargs = dict(
    header=0,
    skiprows=[1],
    converters={
        "SID": str,
        "SEASON": int,
        "BASIN": str,
        "SUBBASIN": str,
        "LON": float,
        "LAT": float,
    },
)


def load(subset, filename, **kwargs):
    if subset.lower() in ["wmo", "usa", "jtwc"]:
        return offline(subset)
    else:
        return online(subset, filename=filename, **kwargs)


def online(subset, filename=None, **kwargs):
    """
    Downloads and load into the current workspace the specified ibtracs subset from the
    IBTrACS archive online.

    Parameters
    ----------
    subset : str
        IBTrACS subset. Can be one of
        * ACTIVE: TCs currently active
        * ALL: Entire IBTrACS database
        * Specific basins: EP, NA, NI, SA, SI, SP, WP
        * last3years: self-explanatory
        * since1980: Entire IBTrACS database since 1980 (advent of satellite era,
                     considered reliable from then on)

    filename : str, optional
        file to which to save the raw data. None to use a temporary file. Default is
        None

    Returns
    -------
    xarray.DataArray
        the IBTrACS subset requested
    """
    # Put IBTrACS specific arguments to read_csv second, so it
    # overwrites any arguments passed
    kwargs = {**kwargs, **online_default_kwargs}

    url = (
        "https://www.ncei.noaa.gov/data/"
        "international-best-track-archive-for-climate-stewardship-ibtracs/"
        f"v04r01/access/csv/ibtracs.{subset}.list.v04r01.csv"
    )

    # filename=None downloads the data to a temporary file
    # Ruff (Flake8 bandit) complains that this url isn't checked, but it explicitly has
    # "https:/" at the start anyway
    filename, _ = urlretrieve(url, filename)  # noqa: S310

    return _csv.load(filename, **kwargs)


def offline(subset="wmo"):
    """
    Function to load offline IBTrACS datasets. These are embedded within the package,
    rather than downloaded online. Because of the offline nature of this feature, the
    datasets might be outdated depending on your last update of the package and/or the
    last update of the datasets by the developers. Last update by the developers for
    this version dates back to May 24th, 2024.

    In order to reduce the size of the embedded files, filtering was made both on the
    attributes and on the records. All offline datasets are based on the "since1980"
    subset of IBTrACS. Recent seasons with tracks marked as "provisionnal" were removed.
    All spur tracks were removed. Only 6-hourly data is provided.

    Two subsets are currently available:
        * "wmo" contains the data provided in the "wmo" columns, which correspond to the
          data provided by the center responsible for the area of a given point
          (see https://community.wmo.int/en/tropical-cyclone-regional-bodies). Note that
          within this dataset, wind units are not homogeneous: they are provided as
          collected from the meteorological agencies, which means that they have
          different time-averaging for wind extrema.
        * "jtwc" contains the data provided in the "wmo" columns, which is provided by
          the NHC or the JTWC.

    For more information, you may see the IBTrACS column documentation at
    https://www.ncei.noaa.gov/sites/default/files/2021-07/IBTrACS_v04_column_documentation.pdf

    Parameters
    ----------
    subset : str, optional
        "jtwc" or "wmo". The default is "wmo".

    Returns
    -------
    xr.Dataset
        The set of tracks requested.

    """
    warnings.warn(
        "This offline function loads a light version of IBTrACS which is embedded"
        " within the package, based on a file produced manually by the developers.\n"
        " It was last updated on the 15th Nov 2024, based on the IBTrACS file at that"
        " date.\n It contains only data from 1980 up to the last year with no"
        " provisional tracks. All spur tracks were removed. Only 6-hourly time steps"
        " were kept."
    )
    if subset.lower() == "wmo":
        warnings.warn(
            "You are loading the IBTrACS-WMO subset. This dataset contains the"
            " positions and intensity reported by the WMO agency responsible for"
            " each basin\n Be aware of the fact that wind and pressure data is provided"
            " as they are in IBTrACS, which means in particular that wind speeds are in"
            " knots and averaged over different time periods.\n For more information,"
            " see the IBTrACS column documentation at"
            " https://www.ncei.noaa.gov/sites/default/files/2021-07/IBTrACS_v04_column_documentation.pdf"
        )
        return _csv.load(wmo_file)
    if subset.lower() in ["usa", "jtwc"]:
        return _csv.load(jtwc_file)


# TODOS:
# Make warnings better
# Deal with units, in general



"""
Module containing function to compute track densities
"""

import numpy as np
import xarray as xr
from scipy.stats import gaussian_kde

import warnings


def density(lon, lat, method="histogram", bin_size=5, crop=False, function_kws=dict()):
    """Function to compute the track density, based on a simple 2D histogram.

    Parameters
    ----------
    lon : array_like
        longitude series
    lat : array_like
        latitude series
    method : str, default="histogram"
        The method used to calculate the density, currently "histogram" or "kde", which
        gives a 2d histogram using `np.histogram2d`
    bin_size : int or float, default=5
        When using histogram, defines the size (in degrees) of the bins.
    n_seasons : int, optional
        Number of season (will be used to divide the final results, so that is
        represents points/year). The default is 1.

    Raises
    ------
    NotImplementedError
        If method given is not 'histogram' or 'kde'

    Returns
    -------
    xarray.DataArray
        Track density as a 2D map.

    """
    # Define coordinates for mapping
    if lon.min() < 0:
        lon_range = (-180, 180)
    else:
        lon_range = (0, 360)
    x_edge = np.arange(lon_range[0], lon_range[1] + bin_size, bin_size)
    y_edge = np.arange(-90, 90 + bin_size, bin_size)
    x_mid, y_mid = (x_edge[1:] + x_edge[:-1]) / 2, (y_edge[1:] + y_edge[:-1]) / 2

    # Compute density
    if method == "histogram":
        h = _histogram(lon, lat, x_edge, y_edge, function_kws)
    elif method == "kde":
        h = _kde(lon, lat, x_mid, y_mid, function_kws)
    else:
        raise NotImplementedError(
            f"Method {method} not implemented yet. Use one 'histogram', 'kde'"
        )

    # Turn into xarray
    da = xr.DataArray(
        h,
        dims=["lat", "lon"],
        coords={"lon": x_mid, "lat": y_mid},
    )

    if crop:  # Crop the map to where there are non-zero points
        da = da.where(da > 0)
        return da.where(~np.isnan(da), drop=True).fillna(0)
    else:
        return da


def _histogram(lon, lat, x_edge, y_edge, function_kws):
    # Compute 2D histogram with numpy
    h, _x, _y = np.histogram2d(lon, lat, bins=[x_edge, y_edge], **function_kws)
    return h.T  # Transpose result


def _kde(lon, lat, x_mid, y_mid, function_kws):
    # engineer positions array for kernel estimation computation
    positions = np.reshape(np.meshgrid(x_mid, y_mid), (2, len(x_mid) * len(y_mid)))
    # Compute kernel density estimate
    kernel = gaussian_kde([lon, lat], **function_kws)
    # Evaluation kernel along positions
    h = np.reshape(kernel(positions), (len(y_mid), len(x_mid)))
    # Account for cell area differences
    warnings.warn(
        "The kde function does not currently take into account the spherical "
        "geometry of the Earth."
    )
    # Normalize so that H integrates to the total number of points
    return h * len(lon) / h.sum()



import functools

from metpy.units import units
import pint
import xarray as xr


def dequantify_results(original_function):
    @functools.wraps(original_function)
    def wrapped_function(*args, **kwargs):
        result = original_function(*args, **kwargs)

        if isinstance(result, tuple):
            return tuple(_dequantify_result(r) for r in result)
        else:
            return _dequantify_result(result)

    return wrapped_function


def _dequantify_result(result):
    if isinstance(result, xr.DataArray) and isinstance(result.data, pint.Quantity):
        return result.metpy.dequantify()
    elif isinstance(result, pint.Quantity) and result.unitless:
        return result.magnitude
    else:
        return result


def validate_units(variable, expected_units):
    if not isinstance(variable, pint.Quantity) or variable.unitless:
        if callable(expected_units):
            expected_units = expected_units(variable)
        variable = variable * units(expected_units)

    return variable



import numpy as np
import pytest

import huracanpy


# TODO properly test tracks1_is_ref with a set of tracks that will give a different
#  answer
@pytest.mark.parametrize("tracks1_is_ref", [True, False])
def test_match(tracks1_is_ref):
    with (
        pytest.warns(
            UserWarning, match="This offline function loads a light version of IBTrACS"
        ),
        pytest.warns(UserWarning, match="You are loading the IBTrACS-WMO subset"),
    ):
        ib = huracanpy.load(source="ibtracs", ibtracs_subset="wmo")
    ref_1996 = ib.where(ib.time.dt.year == 1996, drop=True)
    uz = huracanpy.load(huracanpy.example_year_file)
    uz1 = uz.where(uz.track_id.isin([1207, 1208, 1210, 1212, 1220, 1238]), drop=True)
    uz2 = uz.where(uz.track_id.isin([1207, 1208, 1209, 1211, 1220, 1238]), drop=True)
    matches = huracanpy.assess.match(
        [uz1, uz2, ref_1996], ["UZ1", "UZ2", "ib"], tracks1_is_ref=tracks1_is_ref
    )

    data1, data2, data3 = uz1, uz2, ref_1996
    n1 = len(np.unique(data1.track_id.values))  # Number of tracks in dataset 1
    n2 = len(np.unique(data2.track_id.values))  # Number of tracks in dataset 2
    n3 = len(np.unique(data3.track_id.values))  # Number of tracks in dataset 3

    m_not1 = len(matches[matches.iloc[:, 0].isna()])
    m_not2 = len(matches[matches.iloc[:, 1].isna()])
    m_not3 = len(matches[matches.iloc[:, 2].isna()])
    m_all = len(matches[matches.isna().sum(axis=1) == 0])

    assert len(matches) == 6
    assert (n1, n2, n3) == (6, 6, 118)
    assert (m_not1, m_not2, m_not3, m_all) == (1, 1, 1, 3)


def test_match_pair_empty(tracks_csv, tracks_year):
    matches = huracanpy.assess.match([tracks_csv, tracks_year], ["a", "b"])
    assert matches.size == 0


def test_match_multiple_empty(tracks_year):
    tracks = [
        tracks_year.hrcn.sel_id(track_ids)
        for track_ids in [
            list(range(1207, 1217)),
            list(range(1217, 1227)),
            list(range(1227, 1237)),
        ]
    ]
    with pytest.raises(
        NotImplementedError,
        match="For the moment, the case where two datasets have no match is not",
    ):
        huracanpy.assess.match(tracks, ["a", "b", "c"])


@pytest.mark.parametrize(
    "tracksets, message",
    [([], "You must provide at least two"), ([1, 2, 3], "Number of names provided")],
)
def test_match_fails(tracksets, message):
    with pytest.raises(ValueError, match=message):
        huracanpy.assess.match(tracksets)


def test_scores():
    with (
        pytest.warns(
            UserWarning, match="This offline function loads a light version of IBTrACS"
        ),
        pytest.warns(UserWarning, match="You are loading the IBTrACS-WMO subset"),
    ):
        ib = huracanpy.load(source="ibtracs", ibtracs_subset="wmo")
    ref_1996 = ib.where(ib.time.dt.year == 1996, drop=True)
    uz = huracanpy.load(huracanpy.example_year_file)
    matches = huracanpy.assess.match([uz, ref_1996], ["UZ", "ib"])

    pod = huracanpy.assess.pod(matches, ref_1996, ref_name="ib")
    far = huracanpy.assess.far(matches, uz, detected_name="UZ")

    assert 0.63 < pod < 0.64
    assert 0.14 < far < 0.15



from io import StringIO

from . import _csv


def load(filename):
    with open(filename, "rt") as f:
        line = ""
        while "DATE" not in line:
            line = f.readline().strip()

        # Change first two named columns DATE->time and INDEX->track_id
        # Read others as is (but lower case)
        varnames = ["time", "track_id"] + [var.lower() for var in line.split()][2:]

        # Reformat the file in place like a CSV
        # Header line
        output = [",".join(varnames)]

        # Skip track header lines and collect data as CSV
        output += [",".join(line.split()) for line in f if not line.startswith("Event")]

    # Use existing CSV load function
    return _csv.load(StringIO("\n".join(output)), index_col=False)



"""huracanpy module for saving tracks data"""

from . import _netcdf


def save(dataset, filename):
    """
    Save dataset as filename.
    The file type (NetCDF or csv supported) is detected based on filename extension.

    Parameters
    ----------
    dataset : xarray.Dataset
    filename : str
        Must end in ".nc" or ".csv"

    """
    if filename.split(".")[-1] == "nc":
        _netcdf.save(dataset, filename)
    elif filename.split(".")[-1] == "csv":
        dataset.to_dataframe().to_csv(filename, index=False)
    else:
        raise NotImplementedError(
            "File format not recognized. Please use one of {.nc, .csv}"
        )



"""
Module to load tracks stored as csv files, including TempestExtremes output.
"""

import pandas as pd

# All values recognised as NaN by pandas.read_csv, except "NA" which we want to load
# normally because it is a basin, and added "" to interpret empty entries as NaN
pandas_na_values = [
    " ",
    "#N/A",
    "#N/A N/A",
    "#NA",
    "-1.#IND",
    "-1.#QNAN",
    "-NaN",
    "-nan",
    "1.#IND",
    "1.#QNAN",
    "<NA>",
    "N/A",
    "NULL",
    "NaN",
    "None",
    "n/a",
    "nan",
    "null ",
    "",
]


def load(
    filename,
    load_function=pd.read_csv,
    **kwargs,
):
    """Load csv tracks data as an xarray.Dataset
    These tracks may come from TempestExtremes StitchNodes, or any other source.

    Parameters
    ----------
    filename : str
        The file must contain at least longitude, latitude, time and track ID.
            - longitude and latitude can be named that, or lon and lat.
            - time must be defined a single `time`column or by four columns:
              year, month, day, hour
            - track ID must be within a column named track_id.

    load_function : callable
        One of the load functions in pandas

    **kwargs
        Remaining keywords are passed to the pandas

    Returns
    -------
    xarray.Dataset
    """
    # Update keywords with extra defaults for dealing with "NA" as basin not nan
    # Put kwargs second in this statement, so it can override defaults
    if load_function is pd.read_csv:
        kwargs = {**dict(na_values=pandas_na_values, keep_default_na=False), **kwargs}

    ## Read file
    tracks = load_function(filename, **kwargs)
    # Remove leading/trailing spaces and make all column names lowercase
    tracks.columns = tracks.columns.str.strip().str.lower()

    # Output xr dataset
    tracks = tracks.to_xarray().rename({"index": "record"}).drop_vars("record")

    return tracks



import numpy as np


def inferred_track_id(*variables):
    """Create a track_id variable by combining multiple identifying variables

    If, for example, we have a set of tracks with variables `year` and `storm_number`,
    but `storm_number` is reset to zero for each year, we can still uniquely identify
    each track using the combination of `year` and `storm_number`. This function
    does that identification and creates a track_id array, e.g.

    >>> track_id = huracanpy.inferred_track_id(tracks.year,tracks.storm_number)

    Parameters
    ----------
    *variables

    Returns
    -------
    array_like
        An array of integers ranging from 0 to the total number of tracks. Identifying
        the track of each record. The return type is the same type as the input or a
        numpy.ndarray if the input does not have a `copy` method

    """
    long_id = ""
    for variable in variables:
        _, partial_id = np.unique(variable, return_inverse=True)

        # Split up the individual IDs to avoid accidental matches
        # e.g. (11, 1) and (1, 11) are 11-1 and 1-11 rather than both 111
        long_id = np.char.add(long_id, "-")
        long_id = np.char.add(long_id, partial_id.astype(str))

    _, track_id = np.unique(long_id, return_inverse=True)

    # Try to return the same type as the input
    try:
        track_id_return = variables[0].copy()
        track_id_return[:] = track_id

        return track_id_return

    except AttributeError:
        return track_id



import warnings

import pandas as pd


def parse_any_line(line):
    line_number = line[:5]
    content = line[6:]
    return line_number, content


def identify_line_type(content):
    if len(content) == 28:
        return "track_header"
    elif len(content) == 38:
        return "track_point"
    elif len(content) == 3:
        return "track_type"
    else:
        warnings.warn("Line type not recognized")
        return None


def parse_track_header(content):
    date = content[:10]
    track_length = int(content[13:15])
    track_id = int(content[-4:])
    return date, track_length, track_id


def parse_track_point(content):
    time = content[:13]
    lat = int(content[14:17]) / 10
    lon = int(content[17:21]) / 10
    wind = int(content[22:25])
    pres = int(content[26:30])
    lat_wind = int(content[31:34]) / 10
    lon_wind = int(content[34:38]) / 10
    return time, lat, lon, wind, pres, lat_wind, lon_wind


def load(filename):
    with open(filename, "rt") as f:
        # Read file
        lines = f.readlines()
        # Remove escape character
        lines = [line[:-1] if line.endswith("\n") else line for line in lines]

    # Parse through file
    c, track_length = 0, 0
    lines_out = []
    while len(lines) > 0:
        line = lines.pop(0)
        nb, content = parse_any_line(line)
        line_type = identify_line_type(content)
        if line_type == "track_header":
            # Check that previous track was finished
            if c != track_length:
                raise ValueError("Previous track not finished")
            # Start new track
            c = 0
            date, track_length, track_id = parse_track_header(content)
        elif line_type == "track_point":
            c += 1  # Count point
            time, lat, lon, wind, pres, lat_wind, lon_wind = parse_track_point(content)
            lines_out.append([track_id, time, lat, lon, wind, pres, lat_wind, lon_wind])
        elif line_type == "track_type":
            pass

    # Format
    df = pd.DataFrame(
        lines_out,
        columns=[
            "track_id",
            "time",
            "lat",
            "lon",
            "wind",
            "pres",
            "lat_wind",
            "lon_wind",
        ],
    )
    df["time"] = pd.to_datetime(df.time)

    # Return as xarray
    return df.to_xarray().rename({"index": "record"}).drop_vars("record")



"""Module with function to interpolate the tracks"""

from tqdm import tqdm
import numpy as np
import pandas as pd
import xarray as xr

from ._subset import sel_id


def interp_time(
    tracks,
    track_ids,
    freq="1h",
    prog_bar=False,
):
    """
    Function to interpolate track data at a given frequency.

    Parameters
    ----------
    tracks : xarray.Dataset
        The tracks dataset.
    track_ids : xarray.DataArray
    freq : str, optional
        Frequency at which you want to interpolate the data. The default is '1h'.
    prog_bar : bool, optional

    Returns
    -------
    xarray.Dataset
        The input `xarray.Dataset` with each individual track interpolated to the
        requested frequency
    """

    new_tracks = []

    if prog_bar:
        iterator = tqdm(np.unique(track_ids.values))
    else:
        iterator = np.unique(track_ids.values)

    for tid in iterator:
        t = sel_id(tracks, track_ids, tid)
        new_tracks.append(
            t.set_coords("time")
            .swap_dims({"record": "time"})
            .interp(
                time=pd.date_range(t.time.min().values, t.time.max().values, freq=freq)
            )
            .swap_dims({"time": "record"})
            .reset_coords("time")
        )
    return xr.concat(new_tracks, dim="record")



"""
Functions to plot track/genesis/whatever density

To compute the density, see huracanpy.diags.track_density
"""

import matplotlib.pyplot as plt
import cartopy.crs as ccrs
from cartopy.mpl.geoaxes import GeoAxes


def density(
    d,
    ax=None,
    contourf_kws=dict(cmap="magma_r", levels=10),
    subplot_kws=dict(projection=ccrs.PlateCarree(180)),
    fig_kws=dict(),
    cbar_kwargs={"label": ""},
):
    """Create a map showing the input density

    Parameters
    ----------
    d : xarray.Dataset
        Density map from :func:`huracanpy.calc.density`
    ax :  matplotlib.axes.Axes, optional
        The axes to draw the figure on. A new figure is created if ax is None
    contourf_kws : dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.contourf`
    subplot_kws : dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.subplots` subplot_kw argument
    fig_kws : dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.subplots`
    cbar_kwargs: dict, optional
        Arguments to be passed to :func:`matplotlib.pyplot.colorbar`

    Returns
    -------
    tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        The figure and axes instances created for the plot

    """
    if ax is None:
        fig, ax = plt.subplots(subplot_kw=subplot_kws, **fig_kws)
    else:
        fig = ax.get_figure()

    if isinstance(ax, GeoAxes):
        if "transform" not in contourf_kws:
            contourf_kws["transform"] = ccrs.PlateCarree()
        ax.coastlines()
        ax.gridlines(draw_labels=True)

    d.plot.contourf(ax=ax, cbar_kwargs=cbar_kwargs, **contourf_kws)

    return fig, ax



"""Huracanpy module for plots"""

__all__ = ["tracks", "density", "fancyline", "venn", "doughnut"]

from ._fancyline import fancyline
from ._tracks import tracks
from ._density import density
from ._doughnut import doughnut
from ._venn import venn



"""Huracanpy module for useful auxilliary functions"""

__all__ = [
    "hemisphere",
    "basin",
    "country",
    "continent",
    "is_land",
    "is_ocean",
    "category",
    "time_components",
    "season",
    "inferred_track_id",
]

from ._geography import (
    hemisphere,
    basin,
    country,
    continent,
    is_land,
    is_ocean,
)
from ._category import category
from ._time import time_components, season
from ._utils import inferred_track_id



"""
Test functions that use tricks to speed up their code produce the same result as the
slower method
"""

import pytest
from haversine import haversine_vector, Unit
import numpy as np
import xarray as xr

import huracanpy


@pytest.mark.parametrize(
    ("tracks",),
    (["tracks_csv"], ["tracks_with_extra_coord"]),
)
def test_accel_sel_id(tracks, request):
    tracks = request.getfixturevalue(tracks)
    result = huracanpy.sel_id(tracks, tracks.track_id, 0)

    expected = tracks.groupby("track_id")[0]

    xr.testing.assert_identical(result, expected)


def test_accel_trackswhere():
    # TODO accelerate trackswhere
    pass


@pytest.mark.parametrize(
    ("tracks",),
    (["tracks_csv"], ["tracks_with_extra_coord"]),
)
def test_accel_get_gen_vals(tracks, request):
    tracks = request.getfixturevalue(tracks)
    result = huracanpy.calc.gen_vals(tracks, tracks.time, tracks.track_id)

    with pytest.warns(UserWarning, match="No index created for dimension track_id"):
        expected = tracks.groupby("track_id").first()

    xr.testing.assert_identical(result, expected)


@pytest.mark.parametrize(
    ("tracks",),
    (["tracks_csv"], ["tracks_with_extra_coord"]),
)
def test_accel_get_apex_vals(tracks, request):
    tracks = request.getfixturevalue(tracks)
    result = huracanpy.calc.apex_vals(tracks, tracks.wind10, tracks.track_id)

    with pytest.warns(UserWarning, match="No index created for dimension track_id"):
        expected = tracks.sortby("wind10", ascending=False).groupby("track_id").first()

    xr.testing.assert_identical(result, expected)


@pytest.mark.parametrize(
    ("tracks",),
    (["tracks_csv"], ["tracks_with_extra_coord"]),
)
def test_accel_get_time_from_genesis(tracks, request):
    tracks = request.getfixturevalue(tracks)
    result = huracanpy.calc.time_from_genesis(tracks.time, tracks.track_id)

    track_groups = tracks.groupby("track_id")
    expected = []
    for track_id, track in track_groups:
        expected.append(track.time - track.time[0])

    expected = xr.concat(expected, dim="record")
    expected = expected.rename("time_from_genesis")

    xr.testing.assert_identical(result, expected)


@pytest.mark.parametrize(
    ("tracks",),
    (["tracks_csv"], ["tracks_with_extra_coord"]),
)
def test_accel_get_time_from_apex(tracks, request):
    tracks = request.getfixturevalue(tracks)
    result = huracanpy.calc.time_from_apex(tracks.time, tracks.track_id, tracks.wind10)

    track_groups = tracks.groupby("track_id")
    expected = []
    for track_id, track in track_groups:
        idx = np.argmax(track.wind10.values)
        expected.append(track.time - track.time[idx])

    expected = xr.concat(expected, dim="record")
    expected = expected.rename("time_from_extremum")

    xr.testing.assert_identical(result, expected)


@pytest.mark.parametrize(
    ("tracks",),
    (["tracks_csv"], ["tracks_with_extra_coord"]),
)
def test_accel_match(tracks, request):
    ref = request.getfixturevalue(tracks)
    tracks = ref.where(ref.track_id < 2, drop=True)
    tracks = tracks.where(tracks.time.dt.hour == 0, drop=True)
    tracks["lon"] = tracks.lon + 0.5
    tracks["lat"] = tracks.lat + 0.5

    result = huracanpy.assess.match([tracks, ref])

    max_dist = 300
    track_id1 = []
    track_id2 = []
    npoints = []
    dist = []

    for track_id, track in tracks.groupby("track_id"):
        for track_id_ref, track_ref in ref.groupby("track_id"):
            # Match times
            track_ = track.where(track.time.isin(track_ref.time), drop=True)

            if len(track_.time) > 0:
                track_ref_ = track_ref.where(track_ref.time.isin(track.time), drop=True)

                yx_track = np.asarray([track_.lat, track_.lon]).T
                yx_ref = np.asarray([track_ref_.lat, track_ref_.lon]).T

                dists = haversine_vector(yx_track, yx_ref, Unit.KILOMETERS)

                matches = dists < max_dist
                if matches.any():
                    track_id1.append(track_id)
                    track_id2.append(track_id_ref)

                    dists_track = dists[matches]
                    npoints.append(len(dists_track))
                    dist.append(np.mean(dists_track))

    np.testing.assert_equal(result.id_1, np.asarray(track_id1))
    np.testing.assert_equal(result.id_2, np.asarray(track_id2))
    np.testing.assert_equal(result.temp, np.asarray(npoints))
    np.testing.assert_allclose(result.dist, np.asarray(dist), rtol=1e-12)


@pytest.mark.parametrize(
    ("tracks",),
    (["tracks_csv"], ["tracks_with_extra_coord"]),
)
def test_accel_overlap(tracks, request):
    ref = request.getfixturevalue(tracks)
    tracks = ref.where(ref.track_id < 2, drop=True)
    tracks = tracks.where(tracks.time.dt.hour == 0, drop=True)
    tracks["lon"] = tracks.lon + 0.5
    tracks["lat"] = tracks.lat + 0.5

    result = huracanpy.assess.overlap(tracks, ref)

    delta_start = []
    delta_end = []

    for n, row in result.iterrows():
        track = tracks.where(tracks.track_id == row.id_1, drop=True)
        track_ref = ref.where(ref.track_id == row.id_2, drop=True)

        delta_start.append((track_ref.time[0] - track.time[0]) / np.timedelta64(1, "D"))
        delta_end.append((track_ref.time[-1] - track.time[-1]) / np.timedelta64(1, "D"))

    np.testing.assert_equal(result.delta_start, np.asarray(delta_start))
    np.testing.assert_equal(result.delta_end, np.asarray(delta_end))



"""
Utils related to geographical attributes
"""

import warnings
from pint.errors import UnitStrippedWarning

import numpy as np
import pandas as pd
from shapely.geometry import Point
import geopandas as gpd
from cartopy.io.shapereader import natural_earth
from metpy.xarray import preprocess_and_wrap
from cartopy.crs import Geodetic, PlateCarree

from .._basins import basins


@preprocess_and_wrap(wrap_like="lat")
def hemisphere(lat):
    """
    Function to detect which hemisphere each point corresponds to.

    Parameters
    ----------
    lat : xarray.DataArray

    Returns
    -------
    xarray.DataArray
        The hemisphere series.
        You can append it to your tracks by running

        >>> tracks["hemisphere"] = get_hemisphere(tracks.lat)
    """

    return np.where(lat >= 0, "N", "S")


def basin(lon, lat, convention="WMO-TC", crs=None):
    """
    Function to determine the basin of each point, according to the selected convention.

    Parameters
    ----------
    lon : xarray.DataArray
        Longitude series
    lat : xarray.DataArray
        Latitude series
    convention : str
        Name of the basin convention you want to use.
            * **WMO-TC** - WMO defined tropical cyclone basins
            * **Sainsbury2022JCLI** - Definitions from
                (https://doi.org/10.1175/JCLI-D-21-0712.1) North Atlantic split up into:

                * Main development region (MDR)
                * Subtropical development region (SUB)
                * Western basin / Caribbean sea (WEST)

            * **Sainsbury2022MWR** - Definitions from
                (https://doi.org/10.1175/MWR-D-22-0111.1). Extratropical transition in
                North Atlantic divided into:

                * Europe
                * NoEurope

    crs : cartopy.crs.CRS, optional
        The coordinate reference system of the lon, lat inputs. The basins are defined
        in PlateCarree (-180, 180), so this will transform lon/lat to this projection
        before checking the basin. If None is given, it will use cartopy.crs.Geodetic
        which is essentially the same, but allows the longitudes to be defined in ranges
        broader than -180, 180

    Returns
    -------
    xarray.DataArray
        The basin series.
        You can append it to your tracks by running tracks["basin"] = get_basin(tracks)
    """
    return _get_natural_earth_feature(
        lon,
        lat,
        feature="basin",
        category="physical",
        name=convention,
        resolution=0,
        crs=crs,
    )


# Running this on lots of tracks was very slow if the file is reopened every time this
# is called
_natural_earth_feature_cache = {
    f"physical_{key}_0_basin": value.rename_axis("basin").reset_index()
    for key, value in basins.items()
}


@preprocess_and_wrap(wrap_like="lon")
def _get_natural_earth_feature(lon, lat, feature, category, name, resolution, crs=None):
    key = f"{category}_{name}_{resolution}_{feature}"
    if key in _natural_earth_feature_cache:
        df = _natural_earth_feature_cache[key]
    else:
        fname = natural_earth(resolution=resolution, category=category, name=name)
        df = gpd.read_file(fname)
        df = df[["geometry", feature]]
        _natural_earth_feature_cache[key] = df

    # The metpy wrapper converting to pint causes errors, but I'm still going to use it
    # because it lets me pass different array_like types for lon/lat without writing
    # our own wrapper. For now, just convert anything not a numpy array to a numpy array
    if not isinstance(lon, np.ndarray):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UnitStrippedWarning)
            lon = np.asarray(lon)
            lat = np.asarray(lat)

    if crs is None:
        crs = Geodetic()
    xyz = PlateCarree().transform_points(crs, lon, lat)

    # Create dataframe of points coordinates
    points = pd.DataFrame(dict(coords=list(xyz[:, :2])))
    # Transform into Points within a GeoDataFrame
    points = gpd.GeoDataFrame(points.coords.apply(Point), geometry="coords", crs=df.crs)

    result = np.asarray(
        gpd.tools.sjoin(df, points, how="right", predicate="contains")[feature]
    ).astype(str)

    # Set "nan" as empty
    result[result == "nan"] = ""

    return result


def is_ocean(lon, lat, resolution="10m", crs=None):
    """
    Detect whether each point is over ocean

    Parameters
    ----------
    lon, lat : float or array_like
    resolution : str
        The resolution of the Land/Sea outlines dataset to use. One of

        * 10m (1:10,000,000)
        * 50m (1:50,000,000)
        * 110m (1:110,000,000)

    crs : cartopy.crs.CRS

    Returns
    -------
    array_like[bool]
        Array of "Land" or "Ocean" for each lon/lat point. Should return the same type
        of array as the input lon/lat, or a length 1 :py:class:`numpy.ndarray` if
        lon/lat are floats
    """
    return (
        _get_natural_earth_feature(
            lon,
            lat,
            feature="featurecla",
            category="physical",
            name="ocean",
            resolution=resolution,
            crs=crs,
        )
        == "Ocean"
    )


def is_land(lon, lat, resolution="10m", crs=None):
    """
    Detect whether each point is over land

    Parameters
    ----------
    lon, lat : float or array_like
    resolution : str
        The resolution of the Land/Sea outlines dataset to use. One of

        * 10m (1:10,000,000)
        * 50m (1:50,000,000)
        * 110m (1:110,000,000)

    crs : cartopy.crs.CRS

    Returns
    -------
    array_like[bool]
        Array of "Land" or "Ocean" for each lon/lat point. Should return the same type
        of array as the input lon/lat, or a length 1 :py:class:`numpy.ndarray` if
        lon/lat are floats
    """
    return (
        _get_natural_earth_feature(
            lon,
            lat,
            feature="featurecla",
            category="physical",
            name="ocean",
            resolution=resolution,
            crs=crs,
        )
        == ""
    )


def country(lon, lat, resolution="10m", crs=None):
    """Detect the country each point is over

    Parameters
    ----------
    lon, lat : float or array_like
    resolution : str
        The resolution of the Land/Sea outlines dataset to use. One of

        * 10m (1:10,000,000)
        * 50m (1:50,000,000)
        * 110m (1:110,000,000)

    crs : cartopy.crs.CRS

    Returns
    -------
    array_like
        Array of country names (or empty string for no country) for each lon/lat point.
        Should return the same type of array as the input lon/lat, or a length 1
        :py:class:`numpy.ndarray` if lon/lat are floats
    """
    return _get_natural_earth_feature(
        lon,
        lat,
        feature="NAME",
        category="cultural",
        name="admin_0_countries",
        resolution=resolution,
        crs=crs,
    )


def continent(lon, lat, resolution="10m", crs=None):
    """Detect the continent each point is over

    Parameters
    ----------
    lon, lat : float or array_like
    resolution : str
        The resolution of the Land/Sea outlines dataset to use. One of

        * 10m (1:10,000,000)
        * 50m (1:50,000,000)
        * 110m (1:110,000,000)

    crs : cartopy.crs.CRS

    Returns
    -------
    array_like
        Array of continent names (or empty string for no continent) for each lon/lat
        point. Should return the same type of array as the input lon/lat, or a length 1
        :py:class:`numpy.ndarray` if lon/lat are floats
    """
    return _get_natural_earth_feature(
        lon,
        lat,
        feature="CONTINENT",
        category="cultural",
        name="admin_0_countries",
        resolution=resolution,
        crs=crs,
    )



"""Overlap function"""

from ._match import match


def overlap(tracks1, tracks2, matches=None):
    """
    Function computing the overlap between matched tracks.

    Parameters
    ----------
    tracks1 : xarray.Dataset
    tracks2 : xarray.Dataset
    matches : pandas.Dataframe
        The output from match_tracks on tracks1 and tracks2.
        If None, match_tracks is run on tracks1 and tracks2.

    Returns
    -------
    pd.Dataframe
        Match dataset with added deltas in days
    """
    if matches is None:
        matches = match([tracks1, tracks2])
    c1, c2 = matches.columns[:2].str.slice(3)
    tracks1, tracks2 = tracks1.to_dataframe(), tracks2.to_dataframe()
    matches = (
        matches.join(
            tracks1.groupby("track_id")[["time"]]
            .min()
            .rename(columns={"time": "tmin_" + c1}),
            on="id_" + c1,
        )
        .join(
            tracks1.groupby("track_id")[["time"]]
            .max()
            .rename(columns={"time": "tmax_" + c1}),
            on="id_" + c1,
        )
        .join(
            tracks2.groupby("track_id")[["time"]]
            .min()
            .rename(columns={"time": "tmin_" + c2}),
            on="id_" + c2,
        )
        .join(
            tracks2.groupby("track_id")[["time"]]
            .max()
            .rename(columns={"time": "tmax_" + c2}),
            on="id_" + c2,
        )
    )

    matches["delta_start"] = matches["tmin_" + c2] - matches["tmin_" + c1]
    matches["delta_end"] = matches["tmax_" + c2] - matches["tmax_" + c1]
    matches["delta_end"] = (
        matches.delta_end.dt.days + matches.delta_end.dt.seconds / 86400
    )
    matches["delta_start"] = (
        matches.delta_start.dt.days + matches.delta_start.dt.seconds / 86400
    )

    return matches[["id_" + c1, "id_" + c2, "temp", "dist", "delta_start", "delta_end"]]



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.patheffects import Stroke
from cartopy.crs import Geodetic
from cartopy.mpl.geoaxes import GeoAxes


def _map_values(values, vmin, vmax, vrange, clip):
    # Allow a single value to be set for all the line segments
    if np.size(values) == 1:
        return values

    if vmin is None:
        vmin = values.min()
    if vmax is None:
        vmax = values.max()

    norm = plt.Normalize(vmin, vmax, clip)
    return vrange[0] + norm(values) * (vrange[1] - vrange[0])


def fancyline(
    x,
    y,
    colors=None,
    vmin=None,
    vmax=None,
    cmap=None,
    clip_colors=False,
    linewidths=None,
    wmin=None,
    wmax=None,
    wrange=(1, 5),
    clip_linewidths=True,
    alphas=None,
    amin=None,
    amax=None,
    arange=(0, 1),
    clip_alphas=True,
    linestyles=None,
    ax=None,
    transform=None,
    autoscale=True,
):
    """
    A line plot of x vs y that can show extra information by having a variable color,
    linewidth, alpha, linestyle, or any combination of the four.

    For variable color set a vmin, vmax, and cmap (similar to pcolormesh/contourf)

    >>> fancyline(x, y, colors=z, vmin=0, vmax=10, cmap="viridis")

    For variable linewidth, set a wmin, wmax, and wrange. Like with vmin/vmax, wmin/wmax
    specify the data values for the smallest and largest linewidth with variables
    outside that range being clipping to wmin/wmax. The wrange specifies the range of
    linewidths that you actually want to display

    >>> fancyline(x, y, linewidths=z, wmin=0, wmax=10, wrange=(1, 5))

    For a variable alpha, set an amin, amax, and arange

    >>> fancyline(x, y, alphas=z, amin=0, amax=10, arange=(0, 1))

    For a variable linestyle, pass an array of linestyles specified as strings. e.g. to
    show whether z is above a threshold

    >>> linestyles = ["--" if value < 5 else "-" for value in z]
    >>> fancyline(x, y, linestyles=linestyles)

    https://matplotlib.org/stable/gallery/lines_bars_and_markers/multicolored_line

    Parameters
    ----------
    x, y : array_like
        The points to plot
    colors : array_like or str, optional
        The data used to set the colors along the line or a string of the color to set
        the line to a uniform color
    vmin, vmax : scalar, optional
        The bounding values of the colors array for the colormap. If not set, will use
        the min and max of the colors array
    cmap : str, optional
        Specify the colormap. Will default to the matplotlib default (viridis)
    clip_colors : bool, default=False
        Clips values outside vmin/vmax to vmin/vmax. The default of False is the
        standard behaviour for matplotlib and sets the colours to the outside colors
        from the colormap
    linewidths : array_like or scalar, optional
        The data used to set the width along the line or a single value for the whole
        line
    wmin, wmax : scalar, optional
        The bounding values for the linewidths. If not set, will use the min and max of
        the linewidths array
    wrange : tuple, default=(1, 5)
        The range of linewidths to map the data to
    clip_linewidths : bool, default=True
        Clips values outside wmin/wmax to wmin/wmax. Unlike clip_colors, this is set to
        True by default because negative linewidths are interpreted as positive and to
        avoid very large linewidths on the figure.
    alphas : array_like or scalar, optional
        The data used to set the alpha along the line or a single value for the whole
        line
    amin, amax : float, optional
        The bounding values for the alphas. If not set, will use the min and max of
        the alphas array
    arange : tuple, default=(0, 1)
        The range of alphas to map the data to
    clip_alphas : bool, default=True
        Clips values outside amin/amax to amin/amax. Unlike clip_colors, this is set to
        True by default because alphas outside the range 0-1 will raise an error
    linestyles : array_like or str, optional
    ax : matplotlib.axes.Axes or cartopy.mpl.geoaxes.GeoAxes, optional
        The axes to add the line to. If not given, it will be added to the most recent
        axes
    transform : cartopy coordinate reference system, optional
        If the axes being drawn on are `GeoAxes` then the data will be converted from
        the data coordinates, given by the transform, to the projection coordinates. If
        a transform is not given, then `cartopy.crs.Geodetic` will be used with default
        arguments
    autoscale : bool, default=True
        Determines whether to call ax.autoscale() after adding the lines to the plot
        because the x/y limits are not automatically adjusted otherwise. If you are
        overlaying the line on an existing plot you may want to avoid doing this by
        setting to False.

    Returns
    -------
    matplotlib.collections.LineCollection:
        The plotted LineCollection. Required as argument to `matplotlib.pyplot.colorbar`
    """
    if ax is None:
        ax = plt.gca()

    # Deal with cartopy transforms
    if isinstance(ax, GeoAxes):
        if transform is None:
            transform = Geodetic()

        xyz = ax.projection.transform_points(transform, x, y)
        x = xyz[:, 0]
        y = xyz[:, 1]

    # Break the xy points up in to line segments
    points = np.asarray([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    # Collect the line segments
    lc = LineCollection(segments, path_effects=[Stroke(capstyle="round")])

    if colors is not None:
        if np.size(colors) == 1:
            lc.set_colors(colors)
        else:
            lc.set_array(colors)
            if vmin is None:
                vmin = colors.min()
            if vmax is None:
                vmax = colors.max()
            lc.set_norm(plt.Normalize(vmin, vmax, clip_colors))
            lc.set_cmap(cmap)

    if linewidths is not None:
        lc.set_linewidth(_map_values(linewidths, wmin, wmax, wrange, clip_linewidths))

    if alphas is not None:
        lc.set_alpha(_map_values(alphas, amin, amax, arange, clip_alphas))

    if linestyles is not None:
        lc.set_linestyle(linestyles)

    # Add the colored line to the existing plot
    ax.add_collection(lc)

    if autoscale:
        ax.autoscale()

    return lc



"""Functions for score computation"""

import numpy as np


def pod(matches, ref, ref_name):
    """Probability of Detection

    Parameters
    ----------
    matches : pandas.DataFrame
        The result from matching tracks to a reference dataset output from
        `huracanpy.assess.match`
    ref : xarray.Dataset
        The original reference dataset before matching
    ref_name : str
        The name of the reference dataset in `matches`

    Returns
    -------
    float

    """
    n_detected = matches["id_" + ref_name].nunique()
    n_total = len(np.unique(ref.track_id.values))
    return n_detected / n_total


def far(matches, detected, detected_name):
    """False Attribution Rate

    Parameters
    ----------
    matches : pandas.Dataframe
        The result from matching tracks to a reference dataset output from
        `huracanpy.assess.match`
    detected : xarray.Dataset
        The original dataset that was being matched to the reference
    detected_name : str
        The name of the original dataset in `matches`

    Returns
    -------
    float

    """
    n_matched = matches["id_" + detected_name].nunique()
    n_total = len(np.unique(detected.track_id.values))
    return 1 - (n_matched / n_total)



import numpy as np
import pytest

import huracanpy
from huracanpy._concat import _reset_track_id


def test_reset_track_id(tracks_year):
    tracks = _reset_track_id(tracks_year, tracks_year.track_id, keep_original=True)

    # Check that the old array has been kept
    np.testing.assert_array_equal(tracks.track_id_original, tracks_year.track_id)

    # Check that the new IDs have been assigned correctly
    _, new_track_ids = np.unique(tracks.track_id_original, return_inverse=True)
    np.testing.assert_array_equal(tracks.track_id, new_track_ids)

    # Check that the numbers are as expected
    assert (
        tracks.track_id.sum()
        == (tracks.track_id_original - tracks.track_id_original[0]).sum()
    )


@pytest.mark.parametrize("prefix", [None, "{}_"])
def test_concat_tracks(tracks_csv, prefix):
    tracks = huracanpy.concat_tracks([tracks_csv, tracks_csv], prefix=prefix)

    assert len(tracks.time) == 198
    assert len(tracks.groupby("track_id")) == 6



"""
Functions to plot the tracks
"""

import matplotlib.pyplot as plt
import seaborn as sns
import cartopy.crs as ccrs
from cartopy.mpl.geoaxes import GeoAxes


def tracks(
    lon,
    lat,
    intensity_var=None,
    ax=None,
    subplot_kws=dict(projection=ccrs.PlateCarree(180)),
    fig_kws=dict(figsize=(10, 10)),
    scatter_kws=dict(palette="turbo", color="k"),
):
    """Create a map of all lon/lat points

    Parameters
    ----------
    lon : array_like
    lat : array_like
    intensity_var : array_like, optional
        Colour the individual points by
    ax :  matplotlib.axes.Axes, optional
        The axes to draw the figure on. A new figure is created if ax is None
    subplot_kws : dict, optional
        Keywords passed to the `subplot_kw` argument of `matplotlib.pyplot.subplots`
    fig_kws : dict, optional
        Keywords passed to `matplotlib.pyplot.subplots`
    scatter_kws : dict, optional
        Keywords passed to `seaborn.scatterplot`

    Returns
    -------
    tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        The figure and axes instances created for the plot

    """
    if ax is None:
        fig, ax = plt.subplots(subplot_kw=subplot_kws, **fig_kws)
    else:
        fig = ax.get_figure()

    if isinstance(ax, GeoAxes):
        if "transform" not in scatter_kws:
            scatter_kws["transform"] = ccrs.PlateCarree()
        ax.coastlines()

    sns.scatterplot(
        x=lon,
        y=lat,
        hue=intensity_var,
        ax=ax,
        **scatter_kws,
    )

    return fig, ax



"""
Module containing functions to compute rates.
"""

import warnings

import numpy as np
from metpy.units import units
from metpy.xarray import preprocess_and_wrap

from .._metpy import dequantify_results


def _dummy_track_id(var):
    warnings.warn(
        "track_id is not provided, all points are considered to come from the sametrack"
    )
    return np.zeros(var.shape)


@dequantify_results
@preprocess_and_wrap(wrap_like="var")
def delta(var, track_ids=None, centering="forward"):
    """Take the differences across var, without including differences between the end
    and start of different tracks

    Parameters
    ----------
    var : xarray.DataArray
    track_ids : array_like, optional
    centering : str, optional

    Returns
    -------
    xarray.DataArray

    """
    # TODO: centered centering

    # Curate input
    # If track_id is not provided, all points are considered to belong to the same track
    if track_ids is None:
        track_ids = _dummy_track_id(var)

    # Check that centering is supported
    if centering not in ["forward", "backward"]:
        raise ValueError("centering must be one of ['forward', 'backward']")

    # Compute delta
    delta = var[1:] - var[:-1]

    # Mask points where track_id changes
    # Multiplying np.nan by an array element gives us the correct type of nan for both
    # np.timedelta and pint.Quantity
    delta[track_ids[1:] != track_ids[:-1]] = np.nan * delta[0]

    # Apply centering
    if centering == "forward":
        delta = np.concatenate([delta, [np.nan * delta[0]]])
    elif centering == "backward":
        delta = np.concatenate([[np.nan * delta[0]], delta])

    # Fix for timedeltas
    if np.issubdtype(delta.magnitude.dtype, np.timedelta64):
        delta = delta / np.timedelta64(1, "s")
        delta = delta.magnitude * units("s")

    return delta


@dequantify_results
@preprocess_and_wrap(wrap_like="var")
def rate(var, time, track_ids=None, centering="forward"):
    """Compute rate of change of var, without including differences between the end
    and start of different tracks

    Parameters
    ----------
    var : xarray.DataArray
    time : xarray.DataArray
    track_ids : array_like, optional
    centering : str, optional

    Returns
    -------
    xarray.DataArray

    """
    # Curate input
    # If track_id is not provided, all points are considered to belong to the same track
    if track_ids is None:
        track_ids = _dummy_track_id(var)

    ## Sort data by track_id and time
    # rate_var, track_ids, time = [a.sortby(time) for a in [rate_var, track_ids, time]]
    # rate_var, time, track_ids = [
    #     a.sortby(track_ids) for a in [rate_var, time, track_ids]
    # ]

    # TODO: If var has units, retrieve those

    # Compute deltas
    dx = delta(var, track_ids, centering=centering)
    dt = delta(time, track_ids, centering=centering)

    return dx / dt



"""
Utils related to time
"""

import warnings

from metpy.xarray import preprocess_and_wrap
import numpy as np
from pint.errors import UnitStrippedWarning
import pandas as pd

from ._geography import hemisphere


def time_components(time, components=("year", "month", "day", "hour")):
    """
    Expand the time variable into year/month/day/hour

    Parameters
    ----------
    time : xr.DataArray
        The time series
    components : iterable[str], optional
        The time components of `time` to return. Can include any valid attribute of
        `xarray.core.accessor_dt.DatetimeAccessor`. Default is year, month, day, hour

    Returns
    -------
    list[xarray.DataArray]
        A DataArray for each requested time component

    """
    return [getattr(time.dt, component) for component in components]


@preprocess_and_wrap(wrap_like="track_id")
def season(track_id, lat, time, convention="tc-short"):
    """Determine the cyclone season for each track

    Parameters
    ----------
    track_id : xarray.DataArray
    lat : xarray.DataArray
    time : xarray.DataArray
    convention : str

        * 'tc-short' : In the Northern hemisphere, the season is the same as calendar
          year. In the southern hemisphere, the season n corresponds to July n-1 to
          June n
        * 'tc-long' : In the Northern hemisphere, the season is the same as calendar
          year. In the southern hemisphere, the season from July n-1 to June n is named
          "(n-1)n"

    Raises
    ------
    NotImplementedError
        If convention given is not 'tc-short' or 'tc-long'

    Returns
    -------
    xarray.DataArray
        The season series. You can append it to your tracks by running
        tracks["season"] = get_season(tracks.track_id, tracks.lat, tracks.time)
    """

    # Derive values
    hemi = hemisphere(lat)

    try:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=UnitStrippedWarning,
                message="The unit of the quantity is stripped",
            )
            time = pd.to_datetime(time)
        year = time.year
        month = time.month
    except TypeError:
        # Fix for cftime
        year = np.asarray([t.year for t in time])
        month = np.asarray([t.month for t in time])

    # Store in a dataframe
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UnitStrippedWarning,
            message="The unit of the quantity is stripped",
        )
        df = pd.DataFrame(
            {"hemi": hemi, "year": year, "month": month, "track_id": track_id}
        )
    # Most frequent year, month and hemisphere for each track
    # Grouping is done to avoid labelling differently points in a track that might cross
    # hemisphere or seasons.
    group = df.groupby("track_id")[["year", "month", "hemi"]].agg(
        lambda x: pd.Series.mode(x)[0]
    )

    # Assign season
    if convention == "tc-short":
        season = np.where(group.hemi == "N", group.year, np.nan)
        season = np.where(
            (group.hemi == "S") & (group.month >= 7), group.year + 1, season
        )
        season = np.where((group.hemi == "S") & (group.month <= 6), group.year, season)
    elif convention == "tc-long":
        season = np.where(group.hemi == "N", group.year.astype(str), np.nan)
        season = np.where(
            (group.hemi == "S") & (group.month >= 7),
            group.year.astype(str) + (group.year + 1).astype(str),
            season,
        )
        season = np.where(
            (group.hemi == "S") & (group.month <= 6),
            (group.year - 1).astype(str) + group.year.astype(str),
            season,
        )
    else:
        raise NotImplementedError("Convention not recognized")

    group["season"] = season
    df = df.merge(group[["season"]], on="track_id")

    return df.season.values



"""
Module containing functions to compute ACE
"""

import warnings

import numpy as np
from numpy.polynomial.polynomial import Polynomial
from pint.errors import UnitStrippedWarning
from metpy.xarray import preprocess_and_wrap
from metpy.units import units

from .._metpy import dequantify_results, validate_units


def ace(
    wind,
    sum_by=None,
    threshold=34 * units("knots"),
    wind_units="m s-1",
):
    r"""Calculate accumulate cyclone energy (ACE)

    .. math:: \mathrm{ACE} =
        10^{-4} \sum v_\mathrm{max}^2 \quad (v_\mathrm{max} \ge 34 \mathrm{kn})

    By default, this function will return the "ACE" for each individual point in `wind`.
    To calculate more useful quantities of ACE, use the `sum_by` keyword.

    For example, to calculate the ACE of each individual track, doing

    >>> ace_by_track = huracanpy.tc.ace(tracks.wind, sum_by=tracks.track_id)

    will return a DataArray with track_id as a coordinate and the sum of ACE for each
    track as the data. Note that this is equivalent to using groupby:

    >>> ace_by_point = huracanpy.tc.ace(tracks.wind)
    >>> ace_by_track = ace_by_point.groupby(tracks.track_id).sum()

    To calculate the average ACE by track, you can do

    >>> ace_by_track_mean = ace_by_track.mean()

    Similarly to calculate a climatological mean ACE by year, run

    >>> climatological_ace = huracanpy.tc.ace(
    >>>    tracks.wind, sum_by=tracks.time.dt.year
    >>> ).mean()

    Parameters
    ----------
    wind : array_like
        Maximum velocity of a tropical cyclone associated with the tracks dataset
    sum_by : array_like
        Variable to take the sum of ACE values across. Must have the same length as wind
    threshold : scalar, default=34 knots
        ACE is set to zero below this threshold wind speed. The default argument is in
        knots. To pass an argument with units, use :py:mod:`metpy.units`, otherwise any
        non-default argument will be assumed to have the units of "wind_units" which is
        "m s-1" by default.
    wind_units : str, default="m s-1"
        If the units of wind are not specified in the attributes then the function will
        assume it is in these units before converting to knots

    Returns
    -------
    array_like
        The ACE for each track in wind

    """

    ace_values = _ace(wind, threshold, wind_units)

    if sum_by is not None:
        ace_values = ace_values.groupby(sum_by).sum()

    return ace_values


@dequantify_results
@preprocess_and_wrap(wrap_like="wind")
def _ace(wind, threshold=34 * units("knots"), wind_units="m s-1"):
    wind = validate_units(wind, wind_units)
    wind = wind.to(units("knots"))

    if threshold is not None:
        threshold = validate_units(threshold, wind_units)

        wind[wind < threshold] = 0 * units("knots")

    ace_values = (wind**2.0) * 1e-4

    return ace_values


def pace(
    pressure,
    wind=None,
    model=None,
    sum_by=None,
    threshold_wind=None,
    threshold_pressure=None,
    wind_units="m s-1",
    pressure_units="hPa",
    **kwargs,
):
    """Calculate a pressure-based accumulated cyclone energy (PACE) for each individual
       point

    PACE is calculated the same way as ACE, but the wind is derived from fitting a
    pressure-wind relationship and calculating wind values from pressure using this fit

    Example
    -------
    This function can be called in two ways

    1. Pass the pressure and wind to fit a pressure-wind relationship to the data and
    then calculate pace from the winds derived from this fit

    >>> pace, pw_model = get_pace(pressure, wind)

    The default model to fit is a quadratic polynomial
    (:py:class:`numpy.polynomial.polynomial.Polynomial` with `deg=2`)

    2. Pass just the pressure and an already fit model to calculate the wind speeds from
    this model

    >>> pace, _ = get_pace(pressure, model=pw_model)

    Parameters
    ----------
    pressure : array_like
    wind : array_like, optional
    model : str, class, or object, optional
    sum_by : array_like
        Variable to take the sum of PACE values across. Must have the same length as
        pressure/wind. For examples, see the documentation for `huracanpy.tc.ace`
    threshold_wind : scalar, optional
    threshold_pressure : scalar, optional
    wind_units : str, default="m s-1"
    pressure_units : str, default="Pa"
    **kwargs

    Returns
    -------
    pace_values : array_like

    model : object

    """
    pace_values, model = _pace(
        pressure,
        wind=wind,
        model=model,
        threshold_wind=threshold_wind,
        threshold_pressure=threshold_pressure,
        wind_units=wind_units,
        pressure_units=pressure_units,
        **kwargs,
    )

    if sum_by is not None:
        pace_values = pace_values.groupby(sum_by).sum()

    return pace_values, model


@dequantify_results
@preprocess_and_wrap(wrap_like=("pressure", None))
def _pace(
    pressure,
    wind=None,
    model=None,
    threshold_wind=None,
    threshold_pressure=None,
    wind_units="m s-1",
    pressure_units="hPa",
    **kwargs,
):
    pressure = validate_units(pressure, pressure_units)
    pressure = pressure.to("hPa")
    if wind is not None:
        wind = validate_units(wind, wind_units)
        wind = wind.to(units("knots"))

    model_wind, model = _pressure_wind_relation(
        pressure, wind=wind, model=model, **kwargs
    )
    pace_values = _ace(model_wind, threshold=threshold_wind, wind_units="knots")

    if threshold_pressure is not None:
        threshold_pressure = validate_units(threshold_pressure, pressure_units)

        pace_values[pressure > threshold_pressure] = 0.0 * pace_values.units

    return pace_values, model


def _pressure_wind_relation(pressure, wind=None, model=None, **kwargs):
    if isinstance(model, str):
        if model.lower() == "z2021":
            model = pw_z2021
        elif model.lower() == "holland":
            model = pw_holland

    elif wind is not None:
        if model is None:
            # Here, we calculate a quadratic P/W fit based off of the "control"
            if "deg" not in kwargs:
                kwargs["deg"] = 2
            model = Polynomial.fit(pressure.magnitude, wind.magnitude, **kwargs)

        else:
            model = model.fit(pressure.magnitude, wind.magnitude, **kwargs)

    elif model is None:
        raise ValueError(
            "Need to specify either wind or model to calculate pressure-wind relation"
        )

    wind_from_fit = _get_wind_from_model(pressure.magnitude, model) * units("knots")

    return wind_from_fit, model


def _get_wind_from_model(pressure, model):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UnitStrippedWarning,
            message="The unit of the quantity is stripped when downcasting to ndarray.",
        )

        result = np.asarray(model(pressure))
    return result


# Pre-determined pressure-wind relationships
# Input pressure in hPa, output wind in knots
_z2021 = Polynomial([1.43290190e01, 5.68356519e-01, -1.05371378e-03])


def pw_z2021(pressure):
    return _z2021(1010.0 - pressure)


def pw_holland(pressure):
    return 2.3 * (1010.0 - pressure) ** 0.76



import xarray as xr
from metpy.units import units
import pandas as pd

from . import (
    plot,
    tc,
    info,
    calc,
    save,
    interp_time,
    sel_id,
    trackswhere,
)


@xr.register_dataarray_accessor("hrcn")
class HuracanPyDataArrayAccessor:
    def __init__(self, dataarray):
        self._dataarray = dataarray.copy()

    def nunique(self):
        """
        Method to count number of unique element in a DataArray

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return pd.Series(self._dataarray).nunique()


@xr.register_dataset_accessor("hrcn")
class HuracanPyDatasetAccessor:
    def __init__(self, dataset):
        self._dataset = dataset

    # %% Save
    def save(self, filename):
        """
        Save dataset as filename.
        The file type (NetCDF or csv supported) is detected based on filename extension.

        Parameters
        ----------
        filename : str
            Must end in ".nc" or ".csv"

        Returns
        -------
        None.

        """

        save(self._dataset, filename)

    def sel_id(self, track_id, track_id_name="track_id"):
        return sel_id(self._dataset, self._dataset[track_id_name], track_id)

    def trackswhere(self, condition, track_id_name="track_id"):
        return trackswhere(self._dataset, self._dataset[track_id_name], condition)

    # %% utils
    # ---- geography
    def get_hemisphere(self, lat_name="lat"):
        return info.hemisphere(self._dataset[lat_name])

    def add_hemisphere(self, lat_name="lat"):
        self._dataset["hemisphere"] = self.get_hemisphere(lat_name=lat_name)
        return self._dataset

    def get_basin(self, lon_name="lon", lat_name="lat", convention="WMO-TC", crs=None):
        return info.basin(
            self._dataset[lon_name],
            self._dataset[lat_name],
            convention=convention,
            crs=crs,
        )

    def add_basin(self, lon_name="lon", lat_name="lat", convention="WMO-TC", crs=None):
        self._dataset["basin"] = self.get_basin(lon_name, lat_name, convention, crs)
        return self._dataset

    def get_is_land(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        return info.is_land(
            self._dataset[lon_name],
            self._dataset[lat_name],
            resolution=resolution,
            crs=crs,
        )

    def add_is_land(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        self._dataset["is_land"] = self.get_is_land(lon_name, lat_name, resolution, crs)
        return self._dataset

    def get_is_ocean(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        return info.is_ocean(
            self._dataset[lon_name],
            self._dataset[lat_name],
            resolution=resolution,
            crs=crs,
        )

    def add_is_ocean(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        self._dataset["is_ocean"] = self.get_is_ocean(
            lon_name, lat_name, resolution, crs
        )
        return self._dataset

    def get_country(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        return info.country(
            self._dataset[lon_name],
            self._dataset[lat_name],
            resolution=resolution,
            crs=crs,
        )

    def add_country(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        self._dataset["country"] = self.get_country(lon_name, lat_name, resolution, crs)
        return self._dataset

    def get_continent(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        return info.continent(
            self._dataset[lon_name],
            self._dataset[lat_name],
            resolution=resolution,
            crs=crs,
        )

    def add_continent(self, lon_name="lon", lat_name="lat", resolution="10m", crs=None):
        self._dataset["continent"] = self.get_continent(
            lon_name, lat_name, resolution, crs
        )
        return self._dataset

    # ---- ACE & PACE
    def get_ace(
        self,
        wind_name="wind",
        sum_by=None,
        threshold=34 * units("knots"),
        wind_units="m s-1",
    ):
        """
        Calculate Accumulated Cyclone Energy (ACE) for each point.
        """
        if sum_by is not None:
            sum_by = self._dataset[sum_by]
        return tc.ace(
            self._dataset[wind_name],
            sum_by=sum_by,
            threshold=threshold,
            wind_units=wind_units,
        )

    def add_ace(
        self, wind_name="wind", threshold=34 * units("knots"), wind_units="m s-1"
    ):
        """
        Add ACE calculation to the dataset.
        """
        self._dataset["ace"] = self.get_ace(
            wind_name, sum_by=None, threshold=threshold, wind_units=wind_units
        )
        return self._dataset

    def get_pace(
        self,
        pressure_name="slp",
        wind_name=None,
        model=None,
        sum_by=None,
        threshold_wind=None,
        threshold_pressure=None,
        wind_units="m s-1",
        **kwargs,
    ):
        """
        Calculate Pressure-based Accumulated Cyclone Energy (PACE) for each point.
        """
        pace_values, model = tc.pace(
            self._dataset[pressure_name],
            wind=self._dataset[wind_name] if wind_name else None,
            model=model,
            sum_by=self._dataset[sum_by] if sum_by else None,
            threshold_wind=threshold_wind,
            threshold_pressure=threshold_pressure,
            wind_units=wind_units,
            **kwargs,
        )
        return pace_values, model

    def add_pace(
        self,
        pressure_name="slp",
        wind_name=None,
        model=None,
        threshold_wind=None,
        threshold_pressure=None,
        wind_units="m s-1",
        **kwargs,
    ):
        """
        Add PACE calculation to the dataset.
        """
        pace_values, model = self.get_pace(
            pressure_name=pressure_name,
            wind_name=wind_name,
            model=model,
            sum_by=None,
            threshold_wind=threshold_wind,
            threshold_pressure=threshold_pressure,
            wind_units=wind_units,
            **kwargs,
        )
        self._dataset["pace"] = pace_values
        return self._dataset, model

    # ---- time
    def get_time_components(self, time_name="time"):
        """
        Expand the time variable into year, month, day, and hour.
        """
        return info.time_components(self._dataset[time_name])

    def add_time_components(self, time_name="time"):
        """
        Add year, month, day, and hour as new variables to the dataset.
        """
        components = self.get_time_components(time_name)

        return xr.merge([self._dataset, *components])

    def get_season(
        self,
        track_id_name="track_id",
        lat_name="lat",
        time_name="time",
        convention="tc-short",
    ):
        """
        Derive the season for each track based on latitude and time.
        """
        return info.season(
            self._dataset[track_id_name],
            self._dataset[lat_name],
            self._dataset[time_name],
            convention=convention,
        )

    def add_season(
        self,
        track_id_name="track_id",
        lat_name="lat",
        time_name="time",
        convention="tc-short",
    ):
        """
        Add the season as a new variable to the dataset.
        """
        self._dataset["season"] = self.get_season(
            track_id_name, lat_name, time_name, convention
        )
        return self._dataset

    # --- utils
    def get_inferred_track_id(self, *variable_names):
        return info.inferred_track_id(*[self._dataset[var] for var in variable_names])

    def add_inferred_track_id(self, *variable_names, track_id_name="track_id"):
        self._dataset[track_id_name] = self.get_inferred_track_id(*variable_names)
        return self._dataset

    # --- category
    def get_category(
        self,
        var_name,
        bins=None,
        labels=None,
        variable_units=None,
    ):
        """
        Calculate a generic category from a variable and a set of thresholds.
        """
        return info.category(
            self._dataset[var_name],
            bins=bins,
            labels=labels,
            variable_units=variable_units,
        )

    def add_category(
        self,
        var_name,
        new_var_name=None,
        bins=None,
        labels=None,
        variable_units=None,
    ):
        """
        Add a generic category to the dataset as a new variable.
        """
        if new_var_name is None:
            new_var_name = f"category_{var_name}"
        self._dataset[new_var_name] = self.get_category(
            var_name,
            bins=bins,
            labels=labels,
            variable_units=variable_units,
        )
        return self._dataset

    def get_saffir_simpson_category(
        self, wind_name="wind", convention="Saffir-Simpson", wind_units="m s-1"
    ):
        """
        Determine the Saffir-Simpson Hurricane Scale (SSHS) category.
        """
        return tc.saffir_simpson_category(
            self._dataset[wind_name], convention=convention, wind_units=wind_units
        )

    def add_saffir_simpson_category(
        self, wind_name="wind", convention="Saffir-Simpson", wind_units="m s-1"
    ):
        """
        Add the SSHS category to the dataset.
        """
        self._dataset["saffir_simpson_category"] = self.get_saffir_simpson_category(
            wind_name, convention, wind_units
        )
        return self._dataset

    def get_pressure_category(
        self, slp_name="slp", convention="Klotzbach", slp_units=None
    ):
        """
        Determine the pressure category based on the selected convention.
        """
        return tc.pressure_category(
            self._dataset[slp_name], convention=convention, slp_units=slp_units
        )

    def add_pressure_category(
        self, slp_name="slp", convention="Klotzbach", slp_units=None
    ):
        """
        Add the pressure category to the dataset.
        """
        self._dataset["pressure_category"] = self.get_pressure_category(
            slp_name, convention, slp_units
        )
        return self._dataset

    def get_beta_drift(self, lat_name="lat", wind_name="wind", rmw_name="rmw"):
        return tc.beta_drift(
            self._dataset[lat_name], self._dataset[wind_name], self._dataset[rmw_name]
        )

    def add_beta_drift(self, lat_name="lat", wind_name="wind", rmw_name="rmw"):
        v_drift, theta_drift = self.get_beta_drift(lat_name, wind_name, rmw_name)

        self._dataset["v_drift"] = v_drift
        self._dataset["theta_drift"] = theta_drift
        return self._dataset

    # ---- translation
    def get_azimuth(
        self,
        lon_name="lon",
        lat_name="lat",
        track_id_name="track_id",
        ellps="WGS84",
    ):
        """
        Compute the azimuth between points along a track
        """
        if track_id_name in list(self._dataset.variables):
            return calc.azimuth(
                self._dataset[lon_name],
                self._dataset[lat_name],
                track_id=self._dataset[track_id_name],
                ellps=ellps,
            )
        if (track_id_name is None) or (
            track_id_name not in list(self._dataset.variables)
        ):
            return calc.azimuth(
                self._dataset[lon_name],
                self._dataset[lat_name],
                track_id=None,
                ellps=ellps,
            )

    def add_azimuth(
        self,
        lon_name="lon",
        lat_name="lat",
        track_id_name="track_id",
        ellps="WGS84",
    ):
        """
        Add the azimuth calculation to the dataset.
        """
        self._dataset["azimuth"] = self.get_azimuth(
            lon_name, lat_name, track_id_name, ellps
        )
        return self._dataset

    def get_distance(
        self,
        lon_name="lon",
        lat_name="lat",
        track_id_name="track_id",
        method="geod",
        ellps="WGS84",
    ):
        """
        Compute the distance between points along a track.
        """
        if track_id_name in list(self._dataset.variables):
            return calc.distance(
                self._dataset[lon_name],
                self._dataset[lat_name],
                track_id=self._dataset[track_id_name],
                method=method,
                ellps=ellps,
            )
        if (track_id_name is None) or (
            track_id_name not in list(self._dataset.variables)
        ):
            return calc.distance(
                self._dataset[lon_name],
                self._dataset[lat_name],
                track_id=None,
                method=method,
                ellps=ellps,
            )

    def add_distance(
        self,
        lon_name="lon",
        lat_name="lat",
        track_id_name="track_id",
        method="geod",
        ellps="WGS84",
    ):
        """
        Add the distance calculation to the dataset.
        """
        self._dataset["distance"] = self.get_distance(
            lon_name, lat_name, track_id_name, method, ellps
        )
        return self._dataset

    def get_translation_speed(
        self,
        lon_name="lon",
        lat_name="lat",
        time_name="time",
        track_id_name="track_id",
        method="geod",
        ellps="WGS84",
    ):
        """
        Compute the translation speed along tracks.
        """
        if track_id_name in list(self._dataset.variables):
            return calc.translation_speed(
                self._dataset[lon_name],
                self._dataset[lat_name],
                self._dataset[time_name],
                track_id=self._dataset[track_id_name],
                method=method,
                ellps=ellps,
            )
        if (track_id_name is None) or (
            track_id_name not in list(self._dataset.variables)
        ):
            return calc.translation_speed(
                self._dataset[lon_name],
                self._dataset[lat_name],
                self._dataset[time_name],
                track_id=None,
                method=method,
                ellps=ellps,
            )

    def add_translation_speed(
        self,
        lon_name="lon",
        lat_name="lat",
        time_name="time",
        track_id_name="track_id",
        method="geod",
        ellps="WGS84",
    ):
        """
        Add the translation speed calculation to the dataset.
        """
        self._dataset["translation_speed"] = self.get_translation_speed(
            lon_name, lat_name, time_name, track_id_name, method, ellps
        )
        return self._dataset

    # ---- rates
    def get_delta(self, var_name="wind10", track_id_name="track_id", **kwargs):
        if track_id_name in list(self._dataset.variables):
            return calc.delta(
                self._dataset[var_name],
                track_ids=self._dataset[track_id_name],
                **kwargs,
            )
        if (track_id_name is None) or (
            track_id_name not in list(self._dataset.variables)
        ):
            return calc.delta(self._dataset[var_name], track_ids=None, **kwargs)

    def add_delta(self, var_name="wind10", track_id_name="track_id", **kwargs):
        """
        Add the distance calculation to the dataset.
        """
        self._dataset["delta_" + var_name] = self.get_delta(
            var_name, track_id_name, **kwargs
        )
        return self._dataset

    def get_rate(
        self, var_name="wind10", time_name="time", track_id_name="track_id", **kwargs
    ):
        if track_id_name in list(self._dataset.variables):
            return calc.rate(
                self._dataset[var_name],
                self._dataset[time_name],
                track_ids=self._dataset[track_id_name],
                **kwargs,
            )
        if (track_id_name is None) or (
            track_id_name not in list(self._dataset.variables)
        ):
            return calc.rate(
                self._dataset[var_name],
                self._dataset[time_name],
                track_ids=None,
                **kwargs,
            )

    def add_rate(
        self, var_name="wind10", time_name="time", track_id_name="track_id", **kwargs
    ):
        self._dataset["rate_" + var_name] = self.get_rate(
            var_name, time_name, track_id_name, **kwargs
        )
        return self._dataset

    # ---- interp
    def interp_time(self, freq="1h", track_id_name="track_id", prog_bar=False):
        """
        Interpolate track data at a given frequency.
        """
        return interp_time(
            self._dataset, self._dataset[track_id_name], freq=freq, prog_bar=prog_bar
        )

    # ---- lifecycle
    def get_time_from_genesis(self, time_name="time", track_id_name="track_id"):
        return calc.time_from_genesis(
            self._dataset[time_name], self._dataset[track_id_name]
        )

    def add_time_from_genesis(self, time_name="time", track_id_name="track_id"):
        self._dataset["time_from_genesis"] = self.get_time_from_genesis(
            time_name, track_id_name
        )
        return self._dataset

    def get_time_from_apex(
        self,
        time_name="time",
        track_id_name="track_id",
        intensity_var_name="wind",
        stat="max",
    ):
        return calc.time_from_apex(
            self._dataset[time_name],
            self._dataset[track_id_name],
            self._dataset[intensity_var_name],
            stat=stat,
        )

    def add_time_from_apex(
        self,
        time_name="time",
        track_id_name="track_id",
        intensity_var_name="wind",
        stat="max",
    ):
        self._dataset["time_from_apex"] = self.get_time_from_apex(
            time_name, track_id_name, intensity_var_name, stat
        )
        return self._dataset

    # %% plot
    def plot_tracks(
        self, lon_name="lon", lat_name="lat", intensity_var_name=None, **kwargs
    ):
        if intensity_var_name in list(self._dataset.variables):
            intensity_var = self._dataset[intensity_var_name]
        else:
            intensity_var = None

        return plot.tracks(
            self._dataset[lon_name], self._dataset[lat_name], intensity_var, **kwargs
        )

    def plot_density(
        self, lon_name="lon", lat_name="lat", density_kws=dict(), **kwargs
    ):
        d = self.get_density(lon_name=lon_name, lat_name=lat_name, **density_kws)
        return plot.density(d, **kwargs)

    def plot_fancyline(
        self,
        lon_name="lon",
        lat_name="lat",
        track_id_name="track_id",
        colors=None,
        linewidths=None,
        alphas=None,
        linestyles=None,
        **kwargs,
    ):
        extra_names = dict(
            colors=colors, linewidths=linewidths, alphas=alphas, linestyles=linestyles
        )
        output = []
        for track_id, track in self._dataset.groupby(track_id_name):
            # Allow the other variables to be passed as variable names or constant
            # strings
            # e.g. colors can be a variable on the track or could just be "red"
            extra_variables = {
                key: (track[name] if name in track else name)
                for key, name in extra_names.items()
            }

            output.append(
                plot.fancyline(
                    track[lon_name], track[lat_name], **extra_variables, **kwargs
                )
            )

        return output

    # %% diags
    # ---- density
    def get_density(self, lon_name="lon", lat_name="lat", method="histogram", **kwargs):
        return calc.density(
            self._dataset[lon_name], self._dataset[lat_name], method=method, **kwargs
        )

    # ---- track stats
    def get_track_duration(self, time_name="time", track_id_name="track_id"):
        return calc.track_duration(
            self._dataset[time_name], self._dataset[track_id_name]
        )

    def get_gen_vals(self, time_name="time", track_id_name="track_id"):
        return calc.gen_vals(
            self._dataset, self._dataset[time_name], self._dataset[track_id_name]
        )

    def get_apex_vals(self, var_name, track_id_name="track_id", stat="max"):
        return calc.apex_vals(
            self._dataset,
            variable=self._dataset[var_name],
            track_id=self._dataset[track_id_name],
            stat=stat,
        )



"""
Utils related to translation distance and time
"""

import warnings

from haversine import haversine_vector
from metpy.units import units
from metpy.xarray import preprocess_and_wrap
import numpy as np
from pint.errors import UnitStrippedWarning
import pyproj


from ._rates import delta, _dummy_track_id
from .._metpy import dequantify_results


def _get_distance_azimuth_geod(lon1, lat1, lon2, lat2, ellps="WGS84"):
    # initialize Geod object
    geodesic = pyproj.Geod(ellps=ellps)

    # Compute distance for all data
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UnitStrippedWarning,
            message="The unit of the quantity is stripped when downcasting to ndarray.",
        )
        fwd_azimuth, back_azimuth, dist = geodesic.inv(lon1, lat1, lon2, lat2)

    return dist, fwd_azimuth


def _get_distance_haversine(lon1, lat1, lon2, lat2):
    # Convert longitudes beyond 180° (necessary for haversine to work)
    lon1 = ((lon1 + 180) % 360) - 180
    lon2 = ((lon2 + 180) % 360) - 180

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UnitStrippedWarning,
            message="The unit of the quantity is stripped when downcasting to ndarray.",
        )
        yx1 = np.asarray([lat1, lon1]).T
        yx2 = np.asarray([lat2, lon2]).T

    return haversine_vector(yx1, yx2, unit="m")


@preprocess_and_wrap(wrap_like="lon")
def azimuth(lon, lat, track_id=None, ellps="WGS84"):
    """Compute azimuth between points using geodesic calculation.

    Parameters
    ----------
    lon : xarray.DataArray
    lat : xarray.DataArray
    track_id : array_like, optional
    ellps : str, optional
        The definition of the globe to use for the geodesic calculation (see
        `pyproj.Geod`). Default is "WGS84".
    Returns
    -------
    xarray.DataArray
        Azimuth in degrees.
            0° corresponds to northward (or stagnating);
            90° corresponds to eastward;
            180° corresponds to southward;
            -90° corresponds to westwards.
    """
    if track_id is None:
        track_id = _dummy_track_id(lon)

    # Compute azimuth
    _, azimuth = _get_distance_azimuth_geod(
        lon[:-1], lat[:-1], lon[1:], lat[1:], ellps=ellps
    )

    # Mask track transition points
    azimuth[track_id[1:] != track_id[:-1]] = np.nan * azimuth[0]
    azimuth = np.concatenate([azimuth, [np.nan * azimuth[0]]])

    return azimuth


@dequantify_results
@preprocess_and_wrap(wrap_like="lon")
def distance(lon, lat, *args, track_id=None, method="geod", ellps="WGS84"):
    """Compute distance between longitude/latitude coordinates using
    geodesic or haversine calculation

    >>> distance(lon, lat, track_id)
    Computes the distance between successive lon, lat points, without including
    differences between the end and start points of different tracks

    >>> distance(lon1, lat1, lon2, lat2)
    Computes the distance between each point in (lon1, lat1) and each point in
    (lon2, lat2)

    Parameters
    ----------
    lon : xarray.DataArray
    lat : xarray.DataArray
    *args : xarray.DataArray
        * 0 arguments. Leave empty to calculate distance between successive points
        * 1 argument, track_id. Same as 0 arguments but inserts NaNs where successive
            points are from different tracks
        * 2 arguments, lon and lat arrays. Calculate distances between two tracks,
            e.g. radius of maximum wind speed, using storm centre locations and maximum
            wind speed locations
    track_id : array_like, optional
    method : str, optional
        The method of computing distances, either geodesic (`"geod"`) or haversine
        (`"haversine"`)
    ellps : str, optional
        The definition of the globe to use for the geodesic calculation (see
        `pyproj.Geod`). Default is "WGS84".

    Returns
    -------
    xarray.DataArray

    """
    # TODO: Provide option for centering forward, backwards, centered

    # Curate input
    # If track_id is not provided, all points are considered to belong to the same track
    if len(args) < 2:
        lon1 = lon[:-1]
        lat1 = lat[:-1]
        lon2 = lon[1:]
        lat2 = lat[1:]

        if len(args) == 1:
            if track_id is None:
                track_id = args[0]
            else:
                raise ValueError(
                    "Distance either takes 2 arrays (lon/lat) or 4 arrays 2x(lon/lat)"
                )

        if track_id is None:
            track_id = _dummy_track_id(lon)

    elif len(args) == 2:
        lon1 = lon
        lat1 = lat
        lon2, lat2 = args

    else:
        raise ValueError(
            "Distance either takes 2 arrays (lon/lat) or 4 arrays 2x(lon/lat)"
        )

    if method == "geod":
        dist, _ = _get_distance_azimuth_geod(lon1, lat1, lon2, lat2, ellps=ellps)
    elif method == "haversine":
        dist = _get_distance_haversine(lon1, lat1, lon2, lat2)
    else:
        raise ValueError(
            f"Method {method} for distance calculation not recognised, use one of"
            f"('geod', 'haversince')"
        )

    if len(args) < 2 and track_id is not None:
        dist[track_id[1:] != track_id[:-1]] = np.nan * dist[0]
        dist = np.concatenate([dist, [np.nan * dist[0]]])

    return dist * units("m")


@dequantify_results
@preprocess_and_wrap(wrap_like="lon")
def translation_speed(lon, lat, time, track_id=None, method="geod", ellps="WGS84"):
    """
    Compute translation speed along tracks

    Parameters
    ----------
    lon : xarray.DataArray
    lat : xarray.DataArray
    time : xarray.DataArray
    track_id : array_like, optional
    method : str, optional
        The method of computing distances, either geodesic (`"geod"`) or haversine
        (`"haversine"`)
    ellps : str, optional
        The definition of the globe to use for the geodesic calculation (see
        `pyproj.Geod`). Default is "WGS84".


    Returns
    -------
    xarray.DataArray

    """
    # TODO: Provide option for centering forward, backwards, centered

    # Curate input
    # If track_id is not provided, all points are considered to belong to the same track
    if track_id is None:
        track_id = _dummy_track_id(lon)

    # Distance between each points
    dx = distance(lon, lat, track_id=track_id, method=method, ellps=ellps)

    # time between each points
    dt = delta(time, track_id)

    return dx / dt



import pathlib
import warnings

import pytest

import nbformat
import nbclient


docs_path = pathlib.Path(__file__).parent / "../docs/"
all_notebooks = [(str(fname),) for fname in docs_path.glob("examples/*.ipynb")] + [
    (str(fname),) for fname in docs_path.glob("user_guide/*.ipynb")
]


@pytest.mark.skipif(
    "not config.getoption('--docs')",
    reason="ipynb's are slow to run. Specify --docs explicitly to test execution",
)
@pytest.mark.parametrize(("notebook",), all_notebooks)
def test_notebook_runs(notebook):
    with open(notebook) as f:
        nb = nbformat.read(f, as_version=4)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        nbclient.execute(nb, Xfrozen_modules="off")
    # out = ExecutePreprocessor().preprocess(nb)



import matplotlib.pyplot as plt


def doughnut(values, reference_total, ax=None, **kwargs):
    """Plot a pie chart with a doughnut shape with thickness showing the total number of
    points relative to a reference total

    Based on Figs. 1/2 from
    Roberts et al. (2020) - Impact of Model Resolution on Tropical Cyclone Simulation
    Using the HighResMIP–PRIMAVERA Multimodel Ensemble
    https://doi.org/10.1175/JCLI-D-19-0639.1

    Parameters
    ----------
    values : array_like
        The values for each individual section of the doughnut
    reference_total : scalar
        The value to compare against the total of `values`. If the total of `values` is
        larger than `reference_total` the doughnut will be thicker, and if the total of
        `values` is smaller than `reference_total`, the doughnut will be thinner
    ax : matplotlib.axes.Axes, optional
        Axes to draw the doughnut on. Default will use the most recent axes
    **kwargs
        Remaining arguments are passed to :py:func:`matplotlib.pyplot.pie`

    Returns
    -------
    tuple
        Returns the three variables returned from :py:func:`matplotlib.pyplot.pie`.
        Unlike :py:func:`matplotlib.pyplot.pie`, autotexts will return an empty list if
        no labels are specified, so the length of the tuple is always three
    """
    if ax is None:
        ax = plt.gca()

    ratio = sum(values) / reference_total
    radius = 1 - 1 / (3 * ratio)

    radius = max(radius, 0.05)
    radius = min(radius, 1.0)

    if "wedgeprops" not in kwargs:
        kwargs["wedgeprops"] = dict()
    kwargs["wedgeprops"]["width"] = radius
    if "autopct" in kwargs:
        patches, texts, autotexts = ax.pie(values, **kwargs)
    else:
        patches, texts = ax.pie(values, **kwargs)
        autotexts = []

    return patches, texts, autotexts
